<?php
declare(strict_types=1);

/**
 * webEdition CMS
 *
 * $Rev: 9da22ff058aa $
 * $Author: Lukas Imhof $
 * $Date: 2020-12-03 11:40 +0100 $
 *
 * This source is part of webEdition CMS. webEdition CMS is
 * free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * any later version.
 *
 * The GNU General Public License can be found at
 * http://www.gnu.org/copyleft/gpl.html.
 * A copy is found in the textfile
 * webEdition/licenses/webEditionCMS/License.txt
 *
 * @license	http://www.gnu.org/copyleft/gpl.html  GPL
 */

/**
 * Class we_thumbnail
 *
 * Provides functions for creating and handling webEdition thumbnails.
 */
class we_contents_thumbnail extends we_base_model implements we_exim_xmlIF{
	public const BACKUP_XML_TAG = 'we:thumbnail';

	/** @var string */
	public $ContentType = we_base_ContentTypes::THUMBNAIL; //For backup

	const OK = 0;
	const USE_ORIGINAL = 1;
	const BUILDERROR = 2;
	const NO_GDLIB_ERROR = 3;
	const INPUTFORMAT_NOT_SUPPORTED = 4;
	public const OPTION_CROP = 'Crop';
	public const OPTION_RATIO = 'Ratio';
	public const OPTION_MAXSIZE = 'Maxsize';
	public const OPTION_INTERLACE = 'Interlace';
	public const OPTION_FITINSIDE = 'Fitinside';
	public const OPTION_UNSHARP = 'Unsharp';
	public const OPTION_GAUSSBLUR = 'GaussBlur';
	public const OPTION_NEGATE = 'Negate';
	public const OPTION_GRAY = 'Gray';
	public const OPTION_SEPIA = 'Sepia';
	public const OPTION_DEFAULT = 'default';
	public const FORMAT_PNG = 'png';
	public const FORMAT_JPG = 'jpg';
	public const FORMAT_GIF = 'GIF';

	/** @var int ID of the thumbnail */
	private $thumbID = 0;

	/*	 * @var int Width of the thumbnail */
	private $thumbWidth = 0;

	/** @var int Height of the thumbnail	 */
	private $thumbHeight = 0;

	/** @var int Quality of the jpg thumbnail	 */
	private $thumbQuality = 8;

	/** @var string[] options of the thumbnail	 */
	private $options = [];

	/** @var string Format (jpg, png or gif) of the thumbnail	 */
	private $thumbFormat = '';

	/** @var string Name of the thumbnail	 */
	private $thumbName = '';

	/** @var int ID of the image	 */
	private $imageID = 0;

	/** @var string Filename of the image	 */
	private $imageFileName = '';

	/** @var string Path of the image	 */
	private $imagePath = '';

	/** @var string Extension of the image WITH leading "." */
	private $imageExtension = '';

	/** @var int width of the image	 */
	private $imageWidth = 0;

	/** @var int height of the image	 */
	private $imageHeight = 0;

	/** @var string binaryData of the image (is mostly empty)	 */
	private $imageData = '';

	/** @var ?DateTime date of the thumb last saved in thumbnails table	 */
	private $date;

	/** @var string format (jpg, png or gif) of the generated thumbnail	 */
	private $outputFormat = 'jpg';

	/** @var string path of the generated thumbnail	 */
	private $outputPath = '';

	/** @var int width of the generated thumbnail	 */
	private $outputWidth = 0;

	/** @var int height of the generated thumbnail	 */
	private $outputHeight = 0;

	/** @var bool defines, that even when the original is smaller than desired size, and image should not be maximized, a thumb is generated, needed for Bug #4258: upload of customer image	 */
	private $generateSmaller = false;

	/** @var int[] focus point for resize */
	private $focus = [0, 0];

	public function __construct(){
		parent::__construct(THUMBNAILS_TABLE);
	}

	/**
	 * main initializer
	 *
	 * @return void
	 * @param int $thumbID
	 * @param int $thumbWidth
	 * @param int $thumbHeight
	 * @param string[] $options
	 * @param string $thumbFormat
	 * @param string $thumbName
	 * @param int $imageID
	 * @param string $imageFileName
	 * @param string $imagePath
	 * @param string $imageExtension
	 * @param int $imageWidth
	 * @param int $imageHeight
	 * @param string $imageData
	 * @param ?DateTime $date
	 * @public
	 */
	public function init(int $thumbID, int $thumbWidth, int $thumbHeight, array $options, string $thumbFormat, string $thumbName, int $imageID, string $imageFileName, string $imagePath, string $imageExtension, int $imageWidth, int $imageHeight, string $imageData = '', DateTime $date = null, int $thumbQuality = 8, bool $generateSmaller = false): void{
		$this->thumbID = $thumbID;
		$this->thumbWidth = $thumbWidth;
		$this->thumbHeight = $thumbHeight;
		$this->thumbQuality = $thumbQuality;
		$this->options = array_filter($options);
		$this->thumbFormat = trim($thumbFormat);
		$this->thumbName = $thumbName;
		$this->imageID = $imageID;
		$this->imageFileName = $imageFileName;
		$this->imagePath = $imagePath;
		$this->imageExtension = $imageExtension;
		$this->imageWidth = $imageWidth;
		$this->imageHeight = $imageHeight;
		$this->imageData = $imageData;
		$this->date = $date;
		$this->generateSmaller = $generateSmaller;
		if($this->thumbID && $this->thumbName){
			$this->outputFormat = $this->thumbFormat ?: (array_key_exists(strtolower($this->imageExtension), we_base_imageEdit::GDIMAGE_TYPE) ? we_base_imageEdit::GDIMAGE_TYPE[strtolower($this->imageExtension)] : 'jpg');
			$this->getImageData(false, true);
			$this->checkAndGetImageSizeIfNeeded();
			$this->setOutputPath();
			$this->calculateOutsize();
		}
	}

	/**
	 * initializer if you have all image data and a thumb ID
	 *
	 * @return boolean
	 * @param int $thumbID
	 * @param int $imageID
	 * @param string $imageFileName
	 * @param string $imagePath
	 * @param string $imageExtension
	 * @param int $imageWidth
	 * @param int $imageHeight
	 * @param string $imageData
	 * @public
	 */
	public function initByThumbID(int $thumbID, int $imageID, string $imageFileName, string $imagePath, string $imageExtension, int $imageWidth, int $imageHeight, string $imageData = ''): bool{
		$foo = getHash('SELECT Width,Height,Options,Format,Text,Mod_Date,Quality FROM ' . THUMBNAILS_TABLE . ' WHERE ID=' . $thumbID) ?: [
			'Width' => 0,
			'Height' => 0,
			'Options' => '',
			'Format' => '',
			'Text' => '',
			'Mod_Date' => 'now',
			'Quality' => 0
		];
		$this->init($thumbID, $foo['Width'], $foo['Height'], array_filter(explode(',', $foo['Options'])), $foo['Format'], $foo['Text'], $imageID, $imageFileName, $imagePath, $imageExtension, $imageWidth, $imageHeight, $imageData, new DateTime($foo['Mod_Date']), (int) $foo['Quality']);
		return ($this->thumbID && $this->thumbName);
	}

	/**
	 * initializer if you have all image data and a thumb name
	 *
	 * @return boolean
	 * @param string $thumbName
	 * @param int $imageID
	 * @param string $imageFileName
	 * @param string $imagePath
	 * @param string $imageExtension
	 * @param int $imageWidth
	 * @param int $imageHeight
	 * @param string $imageData
	 * @public
	 */
	public function initByThumbName(string $thumbName, int $imageID, string $imageFileName, string $imagePath, string $imageExtension, int $imageWidth, int $imageHeight, string $imageData = ''): bool{
		$foo = getHash('SELECT ID,Width,Height,Options,Format,Text,Mod_Date,Quality FROM ' . THUMBNAILS_TABLE . ' WHERE Text="' . $this->db->escape($thumbName) . '"') ?: [
			'ID' => 0,
			'Width' => 0,
			'Height' => 0,
			'Options' => '',
			'Format' => '',
			'Text' => '',
			'Mod_Date' => 'now',
			'Quality' => 0
		];
		'@phan-var array{ID:int,Width:int,Height:int,Options:string,Format:string,Text:string,Mod_Date:string,Quality:int} $foo';
		if($foo['ID'] == 0){
			t_e('thumbnail name "' . $thumbName . '" not found.');
		}
		$this->init($foo['ID'], $foo['Width'], $foo['Height'], array_filter(explode(',', $foo['Options'])), $foo['Format'], $foo['Text'], $imageID, $imageFileName, $imagePath, $imageExtension, $imageWidth, $imageHeight, $imageData, new DateTime($foo['Mod_Date']), (int) $foo['Quality']);
		return ($this->thumbID && $this->thumbName);
	}

	/**
	 * initializer if you have only a image ID and a thumb ID
	 *
	 * @return bool
	 * @param int $imageID
	 * @param int $thumbID
	 * @param boolean $getBinary if set, also the binary image data will be loaded
	 * @public
	 */
	public function initByImageIDAndThumbID(int $imageID, int $thumbID, bool $createIfNotExist = true, bool $getBinary = false): bool{
		$this->imageID = $imageID;

		if(!$this->getImageData($getBinary)){
			return false;
		}
		$foo = getHash('SELECT Width,Height,Options,Format,Text,Mod_Date FROM ' . THUMBNAILS_TABLE . ' WHERE ID=' . $thumbID) ?: [
			'Width' => 0,
			'Height' => 0,
			'Options' => '',
			'Format' => '',
			'Text' => '',
			'Mod_Date' => 'now',
		];

		$this->init($thumbID, $foo['Width'], $foo['Height'], array_filter(explode(',', $foo['Options'])), $foo['Format'], $foo['Text'], $imageID, $this->imageFileName, $this->imagePath, $this->imageExtension, $this->imageWidth, $this->imageHeight, $this->imageData, new DateTime($foo['Mod_Date']));

		/* FIXME: the following code was missing here (and in several places where this function is called)!
		 * Is this the right place to execute it? or should we move it to init() or some other place?
		 */
		if(!$this->isImageSvg()){
			try{
				if(($createIfNotExist && (!$this->exists() || (intval(filectime($this->getOutputPath(true))) < $this->getDate()))) && !$this->getThumb()){
					t_e('Error creating thumbnail for file', $this->imageFileName . $this->imageExtension);
					return false;
				}
			} catch (we_exception_base $e){
				t_e($e);
				return false;
			}
		}
		// END

		return true;
	}

	public function getDate(): int{
		return (int) $this->date->format('U') ?: 0;
	}

	/**
	 * creates the thumbnail and saves it in $this->outputPath
	 *
	 * @return string
	 * @public
	 */
	public function getThumb(): string{
		if(we_base_imageEdit::gd_version() <= 0){
			throw new we_exception_image('GD Version unsupported', self::NO_GDLIB_ERROR);
		}
		$ext = $this->imageExtension;
		$type = array_key_exists($ext, we_base_imageEdit::GDIMAGE_TYPE) ? we_base_imageEdit::GDIMAGE_TYPE[$ext] : '';
		if($this->useOriginalSize() && $this->outputFormat == $type){
			//return self::USE_ORIGINAL;
			return WEBEDITION_PATH . '../' . $this->imagePath;
		}

		if(!$type || !we_base_imageEdit::is_imagetype_read_supported($type)){
			throw new we_exception_image('Input format of file is not supported by GDlib (' . $type . '/' . $ext . ')', self::INPUTFORMAT_NOT_SUPPORTED);
		}

		$quality = max(10, min(100, intval($this->thumbQuality) * 10));
		$outarr = we_base_imageEdit::edit_image($this->imageData ?: WEBEDITION_PATH . '../' . $this->imagePath, $this->outputFormat, WEBEDITION_PATH . '../' . $this->outputPath, $quality, $this->thumbWidth, $this->thumbHeight, $this->options, $this->focus, 0);

		if(!$outarr[0]){
			throw new we_exception_image('Error creating thumb (' . $this->imagePath . ',' . $this->outputFormat . ')', self::BUILDERROR);
		}
		return $outarr[0];
	}

	public function getOutputformat(): string{
		return $this->outputFormat;
	}

	public function exists(): bool{
		return !$this->isOriginal() && file_exists($this->getOutputPath(true));
	}

	/**
	 * Gets the Directory for thumbnails
	 *
	 * @static
	 * @public
	 * @return string
	 * @param bool $realpath  if set to true, Document_ROOT will be appended before
	 */
	public static function getThumbDirectory(bool $realpath = false): string{
		return ($realpath ? realpath(WEBEDITION_PATH . '..') : '') . we_base_file::addPreSlash(preg_replace('#^\.?(.*)$#', '${1}', (WE_THUMBNAIL_DIRECTORY ?: '_thumbnails_')));
	}

	/**
	 * function will determine the size of any GIF, JPG, PNG.
	 * This function uses the php Function with the same name.
	 * But the php function doesn't work with some images created from some apps.
	 * So this function uses the gd lib if nothing is returned from the php function
	 *
	 * @static
	 * @public
	 * @return array{int,int,int,string}
	 * @throws we_exception_image
	 * @param string $filename complete path of the image
	 */
	public static function getimagesize(string $filename): array{
		$arr = is_readable($filename) ? getimagesize($filename) : [];

		if(is_array($arr) && (count($arr) >= 4) && $arr[0] && $arr[1]){
			'@phan-var array{int,int,int,string} $arr';
			return $arr;
		}
		if(!we_base_imageEdit::gd_version()){
			throw new we_exception_image('no gd');
		}
		return we_base_imageEdit::getimagesize($filename);
	}

	/**
	 * returns the output path
	 *
	 * @return string
	 * @public
	 */
	public function getOutputPath(bool $withDocumentRoot = false, bool $unique = false): string{
		return ($withDocumentRoot ? realpath(WEBEDITION_PATH . '..') : '') .
			$this->outputPath .
			((!$withDocumentRoot && $unique ) ? '?t=' . ($this->exists() ? filemtime(WEBEDITION_PATH . '..' . $this->outputPath) : time()) :
			'');
	}

	/**
	 * returns the output width
	 *
	 * @return int
	 * @public
	 */
	public function getOutputWidth(): int{
		return we_document_image::isSvg($this->imageExtension) ? $this->thumbWidth : $this->outputWidth;
	}

	/**
	 * returns the output Height
	 *
	 * @return int
	 * @public
	 */
	public function getOutputHeight(): int{
		return we_document_image::isSvg($this->imageExtension) ? $this->thumbHeight : $this->outputHeight;
	}

	// IMI SVG
	/**
	 * returns the output attribute Transform for SVGs
	 *
	 * @return string
	 * @public
	 */
	public function getOutputTransfom(): string{
		return '';
	}

	/**
	 * returns the name of tje thumbnail
	 *
	 * @return string
	 * @public
	 */
	public function getThumbName(): string{
		return $this->thumbName;
	}

	/**
	 * returns true if thumbnail is the same as the original image
	 *
	 * @return boolean
	 * @public
	 */
	public function isOriginal(): bool{
		return $this->outputPath === $this->imagePath;
	}

	/**
	 * returns true if image is svg
	 *
	 * @return boolean
	 * @public
	 */
	public function isImageSvg(): bool{
		return we_document_image::isSvg($this->imageExtension);
	}

	/**
	 * sets the output path for the thumbnail.
	 * if image must not be resized, it will set the path of the original image
	 *
	 * @return void
	 * @private
	 */
	private function setOutputPath(): void{
		if(
			we_base_imageEdit::is_imagetype_supported($this->outputFormat) &&
			( (!$this->useOriginalSize()) || (!$this->hasOriginalType() ) ) &&
			we_base_imageEdit::gd_version() > 0 &&
			we_base_imageEdit::is_imagetype_read_supported(
				array_key_exists(strtolower($this->imageExtension), we_base_imageEdit::GDIMAGE_TYPE) ?
					we_base_imageEdit::GDIMAGE_TYPE[strtolower($this->imageExtension)] : ''
			)
		){
			we_base_file::checkAndMakeFolder(self::getThumbDirectory(true) . '/' . $this->thumbID, true);
			$this->outputPath = self::getThumbDirectory() . '/' . $this->thumbID . '/' . $this->imageID . '_' . $this->imageFileName . '.' . $this->outputFormat;
		} else {
			// IMI SVG
			$this->outputPath = $this->imagePath . we_document_image::isSvg($this->imageExtension) ? $this->getUrlParamsForSvg() : '';
		}
	}

	// IMI SVG
	private function getUrlParamsForSvg(): string{
		return '';
	}

	/**
	 * calculates the real size of the thumbnail (width & height)
	 *
	 * @return void
	 * @private
	 */
	private function calculateOutsize(): void{
		if($this->useOriginalSize()){
			$this->outputWidth = $this->imageWidth;
			$this->outputHeight = $this->imageHeight;
			return;
		}

		$this->outputWidth = 0;
		$this->outputHeight = 0;


		// If width has been specified set it and compute new height based on source area aspect ratio
		if($this->thumbWidth){
			$this->outputWidth = $this->thumbWidth;
			$this->outputHeight = (int) ($this->imageWidth ? round($this->imageHeight * $this->thumbWidth / $this->imageWidth) : 0);
		}

		// If height has been specified set it.
		// If width has already been set and the new image is too tall, compute a new width based
		// on aspect ratio - otherwise, use height and compute new width
		if($this->thumbHeight){
			if($this->outputHeight > $this->thumbHeight || $this->outputHeight == 0){
				$this->outputWidth = (int) ($this->imageHeight ? round($this->imageWidth * $this->thumbHeight / $this->imageHeight) : 0);
				$this->outputHeight = $this->thumbHeight;
			}
		}

		// Check, if we must discard aspect ratio
		// Check if it will fitinside
		if((!in_array(self::OPTION_RATIO, $this->options, true) || in_array(self::OPTION_FITINSIDE, $this->options, true) || in_array(self::OPTION_CROP, $this->options, true)) && ($this->thumbWidth) && ($this->thumbHeight)){
			$this->outputWidth = $this->thumbWidth;
			$this->outputHeight = $this->thumbHeight;
		}
	}

	/**
	 * checks if the thumbnail has the same size as the original image
	 *
	 * @return boolean
	 * @private
	 */
	private function useOriginalSize(): bool{
		return $this->generateSmaller ? false :
			we_base_imageEdit::isNoThumbnailNeeded($this->imageWidth, $this->imageHeight, $this->thumbWidth, $this->thumbHeight, in_array(self::OPTION_MAXSIZE, $this->options, true), in_array(self::OPTION_FITINSIDE, $this->options, true));
	}

	/**
	 * checks if the thumbnail has the same extension as the original image
	 *
	 * @return boolean
	 */
	private function hasOriginalType(): bool{
		return (strtolower($this->imageExtension) === '.' . $this->outputFormat);
	}

	/**
	 * get the image data
	 *
	 * @return boolean
	 * @private
	 */
	private function getImageData(bool $getBinary = false, bool $onlyFocus = false): bool{

		foreach($this->db->getAllq('SELECT c.Name,c.Dat,c.BDID FROM ' . CONTENT_TABLE . ' c WHERE c.DID=' . intval($this->imageID) .
			' AND c.DocumentTable="' . stripTblPrefix(FILE_TABLE) . '"' . ($onlyFocus ? ' AND c.nHash=x\'' . md5('focus') . '\'' : '')) as $item){
			switch($item['Name']){
				case 'origwidth':
					$this->imageWidth = (int) $item['BDID'];
					break;
				case 'origheight':
					$this->imageHeight = (int) $item['BDID'];
					break;
				case 'focus':
					$this->focus = we_unserialize($item['Dat'], [0, 0]);
					break;
			}
		}
		if($onlyFocus){
			return true;
		}

		$imgdat = getHash('SELECT ID,Filename,Extension,Path FROM ' . FILE_TABLE . ' WHERE ID=' . intval($this->imageID));
		if(!$imgdat){
			return false;
		}
		$this->imageFileName = $imgdat['Filename'];
		$this->imagePath = $imgdat['Path'];
		$this->imageExtension = $imgdat['Extension'];

		if($getBinary){
			$this->getBinaryData();
		}

		return true;
	}

	/**
	 * sets width & height of the image if width & height are empty
	 *
	 * @return void
	 * @private
	 */
	private function checkAndGetImageSizeIfNeeded(): void{
		if(!($this->imageWidth && $this->imageHeight)){
			try{
				$arr = self::getimagesize(WEBEDITION_PATH . '../' . $this->imagePath);
				$this->imageWidth = $arr[0];
				$this->imageHeight = $arr[1];
			} catch (we_exception_image $e){

			}
		}
	}

	/**
	 * loads the binary image data
	 *
	 * @return void
	 * @private
	 */
	private function getBinaryData(): void{
		try{
			$this->imageData = we_base_file::load(WEBEDITION_PATH . '../' . $this->imagePath);
		} catch (we_exception_file $e){
			t_e($e);
		}
	}

	public static function deleteByThumbID(int $id): bool{
		$thumbsdir = self::getThumbDirectory(true);
		return we_base_file::deleteLocalFolder($thumbsdir . '/' . $id, true, true);
	}

	public static function deleteByImageID(int $id): bool{
		$thumbsdir = self::getThumbDirectory(true);
		$ret = true;
		foreach(DB_WE::$inst->getAllq('SELECT ID FROM ' . THUMBNAILS_TABLE, true) as $thumbID){
			foreach(glob($thumbsdir . '/' . $thumbID . '/' . $id . '_*', GLOB_NOSORT) as $filename){
				$r = we_base_file::delete($filename);
				$ret = $ret && $r;
			}
		}
		return $ret;
	}

	public static function getJSLangConsts(): string{
		return we_base_util::setLangString([
				'delete_prompt' => g_l('thumbnails', '[delete_prompt]'),
				'empty' => (g_l('alert', '[thumbnail_empty]')),
				'exists' => (g_l('alert', '[thumbnail_exists]')),
				'hochkomma' => (g_l('alert', '[thumbnail_hochkomma]')),
				'new' => g_l('thumbnails', '[new]'),
		]);
	}

	public function toXML(array $attribs = [], array $extraSlots = [], string $content = ''): string{
		return parent::toXML($attribs, ['ContentType' => we_base_request::STRING]);
	}

	public function getPersistent(): array{
		return $this->persistent_slots;
	}

}
