<?php

declare(strict_types=1);

/**
 * webEdition CMS
 *
 * $Rev: a6bd8d02ed4c $
 * $Author: Marc Kraemer <marc.kraemer@webedition.org> $
 * $Date: 2020-12-02 19:49 +0100 $
 *
 * This source is part of webEdition CMS. webEdition CMS is
 * free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * any later version.
 *
 * The GNU General Public License can be found at
 * http://www.gnu.org/copyleft/gpl.html.
 * A copy is found in the textfile
 * webEdition/licenses/webEditionCMS/License.txt
 *
 * @license	http://www.gnu.org/copyleft/gpl.html  GPL
 */
/* a class for handling directories */
abstract class we_folder_base extends we_contents_root{

	const ACTION_DELETE = 'delete';
	const ACTION_UNPUBLISH = 'unpublish';
	const ACTION_PUBLISH = 'publish';
	const ACTION_UNSEARCHABLE = 'unsearchable';
	const ACTION_SEARCHABLE = 'searchable';
	const ACTION_COPY_WS = 'copyws';
	const ACTION_COPY_TID = 'copytid';
	const ACTION_CATEGORY_ADD = 'we_category_new';
	const ACTION_CATEGORY_DELETE = 'we_category_delete';

	/** @var string */
	protected $viewType = 'list';

	/** @var we_doclist_model */
	protected $doclistModel = null;

	/** @var DateTime|null */
	public $Publish_Date = null;

	/** @var int[] */
	protected $Category = [];

	/** @var we_customer_documentFilter	 */
	var $documentCustomerFilter = null;

	/** @var we_search_base */
	var $searchclass;

	/** @var int */
	var $TriggerID = 0;

	/** @var bool */
	public $ApplyWeDocumentCustomerFiltersToChilds = false;

	abstract protected function hasCopyPerm(): bool;

	public function __construct($id = 0){
		//folders are always published
		$this->Publish_Date = new DateTime();
		parent::__construct();
		$this->initByID($id, $this->Table);
	}

	/** initializes all data not static in class and not saved by persistent slots/elements.
	 * This is called from constructor and __wakeup on serialized documents
	 */
	protected function initStatic(): void{
		parent::initStatic();
		$this->persistent_slots = array_merge($this->persistent_slots, [
			'Category' => we_base_request::INTLISTA,
			'Language' => we_base_request::STRING,
			'doclistModel' => we_base_request::STRING,
			'viewType' => we_base_request::STRING
		]);

		if(isWE()){
			$this->EditPageNrs = ($this->ID ?
					array_merge($this->EditPageNrs, [we_base_constants::WE_EDITPAGE_PROPERTIES, we_base_constants::WE_EDITPAGE_INFO]) :
					[we_base_constants::WE_EDITPAGE_PROPERTIES]);


			if(in_array($this->Table, [FILE_TABLE, OBJECT_FILES_TABLE])){
				if(!$this->Language){
					$this->initLanguageFromParent();
				}
				if(we_base_request::bool('we_edit_weDocumentCustomerFilter')){
					$this->documentCustomerFilter = $this->documentCustomerFilter ?: new we_customer_documentFilter($this->Table, $this->ID);
					$this->documentCustomerFilter->getCustomerFilterFromRequest($this->ID);
				}
			}
			$this->adjustEditPageNr();

			if($this->ID){
				if(!$this->doclistModel){
					$this->doclistModel = new we_doclist_model(we_base_WE::$transaction, $this->Table, $this->ID, $this->viewType);
				}
				$this->doclistModel->initByHttp();
			}
		}
	}

	public function we_new(string $table = '', int $parentID = 0, string $name = ''): void{
		$this->ID = 0;
		$this->OldPath = '';
		if($table){
			$this->Table = $table;
			$this->ParentID = $parentID;
			$this->Filename = $name;
			$this->Path = $this->getPath();
			$this->Publish_Date = new DateTime();
		}
		parent::we_new();
		$this->adjustEditPageNr();
	}

	public function getDoclistModel(): ?we_doclist_model{
		return $this->doclistModel ?: ($this->ID ? new we_doclist_model('0', $this->Table, $this->ID, $this->viewType) : null);
	}

	/**
	 * adjust EditPageNrs for CUSTOMERFILTER AND DOCLIST
	 */
	function adjustEditPageNr(): void{
		if(!isWE()){
			return;
		}
		if($this->ID){
			$this->EditPageNrs = array_merge($this->EditPageNrs, [we_base_constants::WE_EDITPAGE_PROPERTIES, we_base_constants::WE_EDITPAGE_INFO]);
		}

		if(we_modules_info::isActive(we_modules_info::CUSTOMER) && we_base_permission::hasPerm([we_base_perm::CUSTOMER_CAN_EDIT_CUSTOMERFILTER, we_base_perm::CUSTOMER_CAN_CHANGE_DOCS])){
			switch($this->Table){
				case FILE_TABLE:
				case OBJECT_FILES_TABLE:
					if($this->ID){
						$this->EditPageNrs = array_merge($this->EditPageNrs, [we_base_constants::WE_EDITPAGE_WEBUSER]);
					}
			}
		}
		switch($this->Table){
			case FILE_TABLE:
			case TEMPLATES_TABLE:
			case VFILE_TABLE:
				if($this->ID){
					$this->EditPageNrs[] = we_base_constants::WE_EDITPAGE_DOCLIST;
				}
		}
	}

	private function initLanguageFromParent(): void{
		$ParentID = $this->ParentID;
		$i = 0;
		for(; !$this->Language; $i++){
			if($ParentID == 0 || $i > 20){
				$this->Language = we_base_WE::$Langs['DefaultFrontendLanguage'];
				if(!$this->Language){
					$this->Language = 'de_DE';
				}
				continue;
			}
			foreach($this->DB_WE->getAllq('SELECT Language,ParentID FROM ' . $this->DB_WE->escape($this->Table) . ' WHERE ID=' . intval($ParentID)) as $rec){
				['ParentID' => $ParentID, 'Language' => $this->Language] = $rec;
			}
		}
	}

	/**
	 * Initialize a folder by its path. Creates the folder and its parents if it does not exist yet.
	 * @param  string  $path          The path to initialize
	 * @param  string  $tblName       The table to store the folder in (default: FILE_TABLE)
	 * @return boolean                Returns true if the folder could be initialized.
	 */
	public function initByPath(string $path, string $tblName = FILE_TABLE): bool{
		// remove tailing '/'
		$path = rtrim($path, '/');

		// Try to retieve the given folder from the database
		$id = fI('SELECT ID FROM ' . $this->DB_WE->escape($tblName) . ' WHERE Path="' . $this->DB_WE->escape($path) . '" AND IsFolder=1');
		if($id){
			// Folder exists: use initByID() to initialize the object.
			$this->initByID($id, $tblName);
		} else {
			// Folder does not exist: create it (if user has permissons to create folders).
			// Iterate the given path and create non-existent folders on-the-fly.
			$spl = explode('/', $path);
			$folderName = (string) array_pop($spl);
			$p = [];
			$last_pid = 0;
			while($spl){
				array_push($p, array_shift($spl));
				$pa = implode('/', $p);
				if($pa){
					if(($pid = fI('SELECT ID FROM ' . $this->DB_WE->escape($tblName) . ' WHERE Path="' . $this->DB_WE->escape($pa) . '"'))){
						$last_pid = $pid;
					} else {
						$folder = self::instance($this->Table);

						$folder->we_new($tblName, $last_pid, end($p) ?: '');
						$folder->save();
						$last_pid = $folder->ID;
					}
				}
			}

			// Eventually initialize the folder object
			$this->we_new($tblName, $last_pid, $folderName);
			$this->save();
		}
		return true;
	}

	protected function isSameAsParent(): bool{
		if($this->ID){
			$pid = $this->ParentID;
			while($pid){
				if($this->ID == $pid){
					return true;
				}
				$pid = fI('SELECT ParentID FROM ' . DB_WE::$inst->escape($this->Table) . '  WHERE ID=' . intval($pid));
			}
		}
		return false;
	}

	protected function checkMovePathExists(): bool{
		return ($this->isMoved() && file_exists($this->OldPath) && file_exists($this->Path));
	}

	/** saves the folder */
	public function we_save(bool $resave = false, bool $skipHook = false): bool{
		//$collectionFolder = $this->Table === VFILE_TABLE;
		//$objFolder = ($this->Table === OBJECT_FILES_TABLE);
		$update = $this->isMoved();
		$tmp = $this->Path;

		if($update){
			//leave old dir for parent save
			$this->Path = $this->OldPath;
		}

		if(!parent::we_save($resave)){
			return false;
		}
		$this->Path = $tmp;

		if(!$this->writeFolder()){
			return false;
		}

		$this->OldPath = $this->Path;
		/* if($this->Table === OBJECT_TABLE){
		  $f = new we_folder_objectFile();
		  $f->initByPath($this->Path, OBJECT_FILES_TABLE, true);
		  } */
		$this->saveDocumentCustomerFilter();

		if(!$resave && $update){
			//FIXME:improve!
			we_navigation_cache::clean(true);
		}

		if(LANGLINK_SUPPORT && $this->Table === FILE_TABLE){
			$this->setLanguageLink();
			//FIXME: what to do if this fails?
		}
		/* hook */
		if(!$skipHook){
			$hook = new we_hook_handler('save', ['obj' => $this, 'resave' => $resave]);
			//check if doc should be saved
			if(!$hook->executeHook()){
				$this->errMsg = $hook->getErrorString();
				return false;
			}

			//check for warnings and notices
			if(($tmp = $hook->getWarningString())){
				$this->warningMsg[] = $tmp;
			}
			if(($tmp = $hook->getNoticeString())){
				$this->noticeMsg[] = $tmp;
			}
		}
		return true;
	}

	protected function saveLangLinkDocuments(bool $withBackLink = true): bool{
		return parent::saveLangLinkDocuments(false);
	}

	protected function setLanguageLinkBacklinks(string $docTable, array $langlinks): void{
		//don't add backlinks
	}

	public static function changeRecursiveRequest(string $what): void{
		$ok = false;
		$lang = 'grant_language';
		$we_transaction = we_base_request::_(we_base_request::TRANSACTION, 'we_cmd', 0, 1);
		$we_doc = we_contents_root::getSessionDocument($we_transaction);
		assert($we_doc instanceof self);
		switch($what){
			case 'we_folder_changeLanguageRecursive':
				$ok = $we_doc->changeLanguageRecursive();
				$lang = '[grant_language]';
				break;
			case 'we_folder_changeTriggerIDRecursive':
				$ok = (we_base_permission::hasPerm([we_base_perm::ADMINISTRATOR]) ?
						$we_doc->changeTriggerIDRecursive() :
						false);
				$lang = '[grant_tid]';
				break;
		}
		$jsCmd = new we_base_jsCmd();
		if($ok){
			$jsCmd->addMsg(g_l('weClass', $lang . '[ok]'), we_base_util::WE_MESSAGE_NOTICE);
		} else {
			$jsCmd->addMsg(g_l('weClass', $lang . '[notok]'), we_base_util::WE_MESSAGE_ERROR);
		}

		echo we_html_tools::getHtmlTop('', $jsCmd->getCmds(), we_html_element::body());
	}

	private function changeLanguageRecursive(): bool{
		$DB_WE = $this->DB_WE;

		$language = $this->Language;
		$documentTable = stripTblPrefix($this->Table);

		// Adapt tblLangLink-entries of documents and objects to the new language
		$docIds = $DB_WE->getAllq('SELECT ID FROM ' . $DB_WE->escape($this->Table) . ' WHERE Path LIKE "' . $DB_WE->escape($this->Path) . '/%" AND Language!="' . $DB_WE->escape($language) . '"', true);
		if(!$docIds){
			return false;
		}
		// First: Change language of published documents, objects
		$DB_WE->query('UPDATE ' . $DB_WE->escape($this->Table) . ' SET ' . we_database_base::arraySetter([
					'Language' => $this->Language
				]) . ' WHERE
	ID IN (' . implode(',', $docIds) . ')');

		//update target lang, if applicable, if not, the old lang is kept. This is ok, since this operation will cause big trouble if tried to change things recursive
		$DB_WE->query('UPDATE IGNORE ' . LANGLINK_TABLE . ' SET ' . we_database_base::arraySetter([
					'Locale' => $this->Language
				]) . ' WHERE DocumentTable="' . $DB_WE->escape($documentTable) . '" AND LDID IN (' . implode(',', $docIds) . ')');
		return true;
	}

	private function changeTriggerIDRecursive(): bool{
		// Change TriggerID of documents
		return $this->DB_WE->query('UPDATE ' . $this->DB_WE->escape($this->Table) . ' SET TriggerID=' . intval($this->TriggerID) . ' WHERE Path LIKE "' . $this->DB_WE->escape($this->Path) . '/%" ContentType IN ("' . implode('","', array_merge(we_base_ContentTypes::FOLDER_TYPES, [
							we_base_ContentTypes::WEDOCUMENT, we_base_ContentTypes::HTML, we_base_ContentTypes::OBJECT_FILE])) . '")');
	}

	/*
	 * not needed if blank and special chars are alloud for collection folders
	  protected function isFilenameNotValid():bool{
	  switch($this->Table){
	  case VFILE_TABLE:
	  $this->Filename = $this->getText();
	  break;
	  default:
	  // do nothing
	  }

	  return parent::isFilenameNotValid();
	  }
	 */

	protected function isFilenameDouble(): bool{
		return fB('SELECT 1 FROM ' . $this->DB_WE->escape($this->Table) . ' WHERE Path="' . $this->DB_WE->escape($this->Path) . '" AND ID!=' . intval($this->ID) . ' LIMIT 1');
	}

	protected function isFilenameEmpty(): bool{
		switch($this->Table){
			case VFILE_TABLE:
				return $this->getText() === '';
			default:
				return $this->Filename === '';
		}
	}

	function userCanSave(bool $ctConditionOk = false): bool{
		switch($this->Table){
			case VFILE_TABLE:
				return we_base_permission::hasPerm([we_base_perm::WORK_COLLECTION_FOLDER_NEW]) && parent::userCanSave(true);
			default:
				return parent::userCanSave($ctConditionOk);
		}
	}

	/* must be called from the editor-script. Returns a filename which has to be included from the global-Script */

	function editor(we_base_jsCmd $jsCmd = null): we_editor_base{
		switch($this->EditPageNr){
			default:
				$_SESSION['weS']['EditPageNr'] = $this->EditPageNr = we_base_constants::WE_EDITPAGE_PROPERTIES;
			case we_base_constants::WE_EDITPAGE_PROPERTIES:
				return new we_editor_properties($this, $jsCmd);
			case we_base_constants::WE_EDITPAGE_INFO:
				return new we_editor_info($this, $jsCmd);
			case we_base_constants::WE_EDITPAGE_WEBUSER:
				return new we_editor_weDocumentCustomerFilter($this, $jsCmd);
			case we_base_constants::WE_EDITPAGE_DOCLIST:
				return new we_editor_doclist($this, $jsCmd);
		}
	}

	protected function userCanChangePath(): bool{
		$userCanChange = we_base_permission::hasPerm([we_base_perm::WORK_DOCUMENT_FOLDER_PATH_CHANGE]) || ($this->CreatorID == we_users_user::getID()) || (!$this->ID);
		if($this->ID != 0 && $this->ParentID == 0 && $this->ParentPath === '/' && $this->Table === OBJECT_FILES_TABLE){
			return false;
		}
		return $userCanChange;
	}

	function formPath(bool $disablePath = false, bool $notSetHot = false, string $extra = ''): string{
		$ws = we_users_util::get_ws($this->Table);
		if(intval($this->ParentID) == 0 && $ws){
			$this->ParentID = (int) reset($ws);
			$this->ParentPath = id2path($this->ParentID, $this->Table);
		}

		return (!$this->userCanChangePath() ? $this->Path :
				$this->formInput('', (in_array($this->Table, [FILE_TABLE, TEMPLATES_TABLE, OBJECT_FILES_TABLE, OBJECT_TABLE, VFILE_TABLE], true)) ? 'Filename' : 'Text', g_l('weClass', '[foldername]'), 0, 255, [
					'onchange' => "pathOfDocumentChanged(true);"]) . we_html_element::br() .
				$this->formDirChooser() . we_html_element::br() .
				($this->Table === OBJECT_FILES_TABLE ?
				$this->formTriggerDocument($this->TriggerID) . we_html_element::br() .
				we_html_tools::alertAttentionBox(g_l('weClass', '[grant_tid][expl]') . ($this->ID ? '' : g_l('weClass', '[availableAfterSave]')), we_html_tools::TYPE_INFO, 0, false) .
				we_html_button::create_button(we_html_button::OK, 'javascript:if(_EditorFrame.getEditorIsHot()) {WE().util.showMessage(WE().consts.g_l.alert.saveFirstMessage,WE().consts.message.WE_MESSAGE_ERROR, window);' . " } else {;we_cmd('we_folder_changeTriggerIDRecursive','" . we_base_WE::$transaction . "');}", '', '', ($this->ID ? false : true)) :
				'')
				);
	}

	function formChangeOwners(): string{
		return we_html_tools::alertAttentionBox(g_l('modules_users', '[grant_owners_expl]') . ($this->ID ? '' : ' ' . g_l('weClass', '[availableAfterSave]')), we_html_tools::TYPE_INFO, 390, false) .
				we_html_button::create_button(we_html_button::OK, 'javascript:if(_EditorFrame.getEditorIsHot()) {WE().util.showMessage(WE().consts.g_l.alert.input_file_name,WE().consts.message.WE_MESSAGE_ERROR, window);' . " } else {;we_cmd('we_users_changeR','" . we_base_WE::$transaction . "');}", '', '', !$this->ID);
	}

	function formChangeLanguage(): string{
		return
				we_html_tools::alertAttentionBox(g_l('weClass', '[grant_language][expl]') . ($this->ID ? '' : ' ' . g_l('weClass', '[availableAfterSave]')), we_html_tools::TYPE_INFO, 390, false) .
				we_html_button::create_button(we_html_button::OK, "javascript:if(_EditorFrame.getEditorIsHot()) { WE().util.showMessage(WE().consts.g_l.alert.input_file_name,WE().consts.message.WE_MESSAGE_ERROR, window); } else {;we_cmd('we_folder_changeLanguageRecursive','" . we_base_WE::$transaction . "');}", '', '', !$this->ID);
	}

	protected function formCopyDocument(): string{
		$idname = 'we_' . $this->Name . '_CopyID';
		$parents = array_merge([0, $this->ID], we_base_util::getParentIDs(FILE_TABLE, $this->ID));

		return we_html_tools::alertAttentionBox(
						g_l('weClass', '[copy_owners_expl]') .
						($this->ID ? '' : ' ' . g_l('weClass', '[availableAfterSave]')
						), we_html_tools::TYPE_INFO, 0, false) .
				we_html_element::hidden($idname, $this->CopyID) .
				we_html_button::create_button(we_html_button::SELECT, ($this->ID ?
						"javascript:we_cmd('we_selector_directory', {id:document.we_form.elements['" . $idname . "'].value, table:'" . $this->Table . "', targetID:'" . $idname . "', jsCmd:['copyFolderCheck," . $this->ID . "," . $this->Table . "," . implode(',', $parents) . "']})" :
						"javascript:WE().util.showMessage(WE().consts.g_l.alert.copy_folders_no_id,WE().consts.message.WE_MESSAGE_ERROR, window);"), '', '', !$this->ID);
	}

	protected function writeFolder(): bool{
		if($this->Path == $this->OldPath || !$this->OldPath){
			return $this->saveToServer();
		}
		if(!$this->moveAtServer()){
			return false;
		}
		$this->modifyLinks();
		$this->modifyChildrenPath();

		$this->OldPath = $this->Path;
		return true;
	}

	protected function modifyLinks(): void{
		switch($this->Table){
			case FILE_TABLE:
			case TEMPLATES_TABLE:
			case OBJECT_FILES_TABLE:
				$this->DB_WE->query('UPDATE ' . $this->DB_WE->escape($this->Table) . ' SET Path=CONCAT("' . $this->DB_WE->escape($this->Path) . '",SUBSTRING(Path,' . (strlen($this->OldPath) + 1) . ')) WHERE Path LIKE "' . $this->DB_WE->escape($this->OldPath) . '/%" OR Path="' . $this->DB_WE->escape($this->OldPath) . '"');
		}
	}

	public function modifyChildrenPath(): void{
		@ignore_user_abort(true);
		$DB_WE = new DB_WE();
		$entries = $this->DB_WE->getAllFirstq('SELECT ID,ContentType FROM ' . $DB_WE->escape($this->Table) . ' WHERE ParentID=' . intval($this->ID), false);
		foreach($entries as $id => $ct){
			if($ct){
				$we_doc = we_base_ContentTypes::getObject((string) $ct);
				assert($we_doc instanceof we_document_base);
				$we_doc->initByID((int) $id, $this->Table, self::LOAD_TEMP_DB);
				$we_doc->ModifyPathInformation($this->ID);
			} else {
				t_e('No class set at entry ', $id, $this->Table);
			}
		}
		@ignore_user_abort(false);
	}

	protected function moveAtServer(): bool{
		return true;
	}

	/**
	 * Create the folders on the server in the public and the site directory.
	 * Note: for internal use
	 *
	 * @return boolean true on success, false in case of failure
	 */
	protected function saveToServer(): bool{
		return true;
	}

	/**
	 * Beseitigt #Bug 3705: sorgt daf�r, das auch leere Dokumentenordner bei einem REbuild angelegt werden
	 */
	public function we_rewrite(): bool{
		return (parent::we_rewrite() ? (in_array($this->Table, [FILE_TABLE, TEMPLATES_TABLE]) ? $this->we_save(true) : true) : false);
	}

	/**
	 * @desc	the function modifies document EditPageNrs set
	 */
	function checkTabs(): void{

	}

	public static function urlReplacePath(string $path, bool $hostMatch = false, int $max = -1, int &$cnt = 0): string{
		static $all = null;
		if($all === null){
			$all = [];
			$tmp = DB_WE::$inst->getAllFirstq('SELECT Path,urlMap FROM ' . FILE_TABLE . ' WHERE urlMap!="" ORDER BY Path DESC', false);
			'@phan-var array<string,string> $tmp';
			$last = '';
			foreach($tmp as $dbpath => $map){
				$host = trim(str_replace(['https://', 'http://'], '', $map), '/');
				if($_SERVER['SERVER_NAME'] === $host){
					$last = '-^' . preg_quote((string) $dbpath, '-') . '(/.*)-';
				}
				$all['-^' . preg_quote((string) $dbpath, '-') . '(/.*)-'] = $host;
			}
			if($last){
				$all[$last] = $_SERVER['SERVER_NAME'];
			}
		}
		if(!$all){
			return $path;
		}
		return preg_replace(array_keys($all), array_map(function($host) use($hostMatch){
					return $hostMatch || $_SERVER['SERVER_NAME'] !== $host ? '//' . $host . '${1}' : '${1}';
				}, $all), $path, $max, $cnt);
	}

	public static function urlReplaceDocument(string $doc, bool $hostMatch = false, int $max = -1, int &$cnt = 0): string{
		//FIXME $hostMatch not used!!
		static $paths = null;
		static $pathsQ = '';
		if($paths === null){
			$paths = [];
			$tmp = DB_WE::$inst->getAllFirstq('SELECT Path,urlMap FROM ' . FILE_TABLE . ' WHERE urlMap!="" ORDER BY Path DESC', false);
			'@phan-var array<string,string> $tmp';
			$last = '';
			foreach($tmp as $path => $map){
				$host = trim((string) str_replace(['https://', 'http://'], '', $map), '/');
				if($_SERVER['SERVER_NAME'] === $host){
					$last = '-^' . preg_quote((string) $path, '-') . '(/.*)-';
				}
				$paths['-^' . preg_quote((string) $path, '-') . '(/.*)-'] = $host;
			}
			if($last){
				$paths[$last] = $_SERVER['SERVER_NAME'];
			}
			$pathsQ = implode('|', array_map(function($s){
						return preg_quote($s, '-');
					}, array_keys($paths)));
		}

		if(!$paths){
			return $doc;
		}

		return preg_replace_callback('-(href|src|action|location|content|srcset|url)(\s*=\s*["\']|\s*\(\s*)' .
				'(|.+,\s*)(' . $pathsQ . ')(/[^"\'\)]*)(["\'\)])-', function($m) use($paths, $pathsQ){
			unset($m[0]);
			$m[4] = strtr($m[4], $paths); //this has matched exactly
			$m[5] = ($m[1] === 'srcset' ? //this attribute has multiple paths
					preg_replace_callback('-(,\s*)(' . $pathsQ . ')-', function($s) use($paths){
						return $s[1] . strtr($s[2], $paths);
					}, $m[5]) :
					$m[5]);
			return implode('', $m);
		}, $doc);
	}

	protected function updateRemoteLang(we_database_base $db, int $id, string $lang, string $type): void{
		$oldLang = fS('SELECT Language FROM ' . $db->escape($this->Table) . ' WHERE ID=' . $id);
		if($oldLang == $lang){
			return;
		}
		//update Lang of doc
		$db->query('UPDATE ' . $db->escape($this->Table) . ' SET Language="' . $db->escape($lang) . '" WHERE ID=' . $id);
		//update LangLink:
		$db->query('UPDATE ' . LANGLINK_TABLE . ' SET DLocale="' . $db->escape($lang) . '" WHERE DID=' . $id . ' AND DocumentTable="' . $db->escape($type) . '"');
		//drop invalid entries => is this safe???
		$db->query('DELETE FROM ' . LANGLINK_TABLE . ' WHERE DID=' . $id . ' AND DocumentTable="' . $db->escape($type) . '" AND DLocale!="' . $db->escape($lang) . '"');
	}

	public static function getUrlReplacements(we_database_base $db, bool $onlyUrl = false, bool $hostMatch = false): array{
		static $ret = -1;
		if($ret == -1){
			$ret = ['full' => [], 'url' => [], 'full_host' => [], 'url_host' => [],];
			$all = DB_WE::$inst->getAllFirstq('SELECT Path,urlMap FROM ' . FILE_TABLE . ' WHERE urlMap!="" ORDER BY Path DESC', false);
			'@phan-var array<string,string> $all';
			$lastRules = [];
			foreach($all as $path => $map){
				$host = trim(str_replace(['https://', 'http://'], '', $map), '/');
				$path = preg_quote($path, '-');
				$rep1 = '-((href\s*=|src\s*=|action\s*=|location\s*=|content\s*=|srcset\s*=|url)\s*["\'\(])(' . $path . ')(/[^"\'\)]*["\'\)])-';
				$rep2 = '-^' . $path . '(/.*)-';
				if($_SERVER['SERVER_NAME'] === $host){
					//this must be at the end, since duplicate replacements may need to match the original string
					$lastRules['full']['${1}${4}'] = $rep1;
					$lastRules['url']['${1}'] = $rep2;
				} else {
					$ret['full']['${1}//' . $host . '${4}'] = $rep1;
					$ret['url']['//' . $host . '${1}'] = $rep2;
				}
				$ret['full_host']['${1}' . '//' . $host . '${4}'] = $rep1;
				$ret['url_host']['//' . $host . '${1}'] = $rep2;
			}
			if($lastRules){
				$ret['full'] = array_merge($ret['full'], $lastRules['full']);
				$ret['url'] = array_merge($ret['url'], $lastRules['url']);
			}
		}
		return $ret[($onlyUrl ? 'url' : 'full') . ($hostMatch ? '_host' : '')];
	}

	public static function getUrlFromID(int $id): string{
		if(!$id){
			return '';
		}
		$replace = self::getUrlReplacements(DB_WE::$inst, true, true);
		$path = fS('SELECT Path FROM ' . FILE_TABLE . ' WHERE ID=' . $id);
		$cnt = 0;
		$ret = $replace ?
				preg_replace($replace, array_keys($replace), $path, 1, $cnt) :
				$path;
		return ($cnt ? we_http_util::getServerProtocol() . ':' : '') . $ret;
	}

	public function getPropertyPage(we_base_jsCmd $jsCmd): string{
		$parts = (new we_html_multiIconBoxContent())
				->add(new we_html_multiIconBoxContentLine(g_l('weClass', '[path]'), $this->formPath(), we_html_multiIconBox::PROP_PATH));

		switch($this->Table){
			case FILE_TABLE:
			case OBJECT_FILES_TABLE:
				if((count(we_base_WE::$Langs['FrontendLanguages']) > 1)){
					$parts->add((new we_html_multiIconBoxContentLine(g_l('weClass', '[language]'), $this->formLangLinks() . '<br/>' .
									we_html_element::div(['class' => "weMultiIconBoxHeadline"], g_l('weClass', '[grant_language][headline]')) .
									$this->formChangeLanguage(), we_html_multiIconBox::PROP_LANG))->fold());
				}
		}
		if($this->hasCopyPerm()){
			$parts->add(new we_html_multiIconBoxContentLine(g_l('weClass', '[copyFolder]'), $this->formCopyDocument(), we_html_multiIconBox::PROP_COPY));
		}

		switch($this->Table){
			case FILE_TABLE:
			case OBJECT_FILES_TABLE :
				$parts->add((new we_html_multiIconBoxContentLine(g_l('weClass', '[owners]'), $this->formCreatorOwners($jsCmd) . '<br/>' .
								(we_base_permission::hasPerm([we_base_perm::ADMINISTRATOR]) ?
										we_html_element::div(['class' => 'weMultiIconBoxHeadline'], g_l('modules_users', '[grant_owners]')) .
										$this->formChangeOwners() : ''), we_html_multiIconBox::PROP_USER))->fold());
		}

		return (new we_html_multiIconBox($parts, 'PropertyPage'))->marginLeft()->get();
	}

	protected function saveContentDataInDB(string $table): bool{
		return true;
	}

	public function getLockedpageNr(): int{
		return in_array(we_base_constants::WE_EDITPAGE_DOCLIST, $this->EditPageNrs, true) ? we_base_constants::WE_EDITPAGE_DOCLIST : we_base_constants::WE_EDITPAGE_FIELDS;
	}

	public static function instance(string $table, int $id = 0): we_folder_base{
		switch($table){
			default:
				throw new we_exception_base('no folder instance for ' . $table);
			case FILE_TABLE:
				return new we_folder_document($id);
			case OBJECT_FILES_TABLE:
				return new we_folder_objectFile($id);
			case TEMPLATES_TABLE:
				return new we_folder_template($id);
			case VFILE_TABLE:
				return new we_folder_vfile($id);
			case NAVIGATION_TABLE:
			case EXPORT_TABLE:
			case NEWSLETTER_TABLE:
			case VOTING_TABLE:
				return new we_folder_module($table, $id);
		}
	}

	/**
	 * inits weDocumentCustomerFilter from db regarding the modelId
	 * is called from "we_textContentDocument::we_load"
	 * @see we_document_textContent::we_load
	 */
	protected function initWeDocumentCustomerFilterFromDB(): void{
		if(we_modules_info::isActive(we_modules_info::CUSTOMER)){
			$this->documentCustomerFilter = we_customer_documentFilter::getFilterByIdAndTable($this->ID, $this->Table);
		}
	}

	function __sleep(): array{
		return array_merge(parent::__sleep(), ['documentCustomerFilter']);
	}

	public function getDocumentCustomerFilter(): ?we_customer_documentFilter{
		return $this->documentCustomerFilter;
	}

	public function setText(string $text): void{
		$this->Filename = $text;
	}

	/**
	 * resave weDocumentCustomerFilter
	 *
	 */
	protected function saveDocumentCustomerFilter(): void{
		if(!we_modules_info::isActive(we_modules_info::CUSTOMER) || empty($this->documentCustomerFilter)){
			return;
		}
		$this->documentCustomerFilter->setModelId($this->ID);
		$this->documentCustomerFilter->save();
	}

	protected function applyWeDocumentCustomerFilterFromFolder(): void{
		if(!we_modules_info::isActive(we_modules_info::CUSTOMER)){
			return;
		}
		$tmpFolder = we_folder_base::instance($this->Table, $this->ParentID);
		$this->documentCustomerFilter = $tmpFolder->documentCustomerFilter;

		$this->ApplyWeDocumentCustomerFiltersToChilds = ($this->ID != 0);
	}

	function publInScheduleTable(): ?DateTime{
		return null;
	}

}
