<?php
declare(strict_types=1);

/** webEdition CMS
 *
 * $Rev: 48ae4a4d17d1 $
 * $Author: Lukas Imhof $
 * $Date: 2020-10-29 12:27 +0100 $
 *
 * This source is part of webEdition CMS. webEdition CMS is
 * free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * any later version.
 *
 * The GNU General Public License can be found at
 * http://www.gnu.org/copyleft/gpl.html.
 * A copy is found in the textfile
 * webEdition/licenses/webEditionCMS/License.txt
 *
 * @license	http://qwww.gnu.org/copyleft/gpl.html  GPL
 */
abstract class we_gui_login{
	/** @var int */
	public static $userLoginDenied = 0;

	/** @var string */
	public static $userLoginDeniedErr = '';

	private static function checkCriticalSettings(): bool{
//Check some critical PHP Setings #7243
//FIXME: implement class sysinfo.class, for not analysing the same php settings twice (here and in sysinfo.php)
		if(we_base_permission::hasPerm([we_base_perm::ADMINISTRATOR])){
			//$suhosinMsg = (extension_loaded('suhosin') && !in_array(ini_get('suhosin.simulation'), ['1', 'on', 'yes', 'true'], true)) ? 'suhosin=on\n' : '';

			$maxInputMsg = (!ini_get('max_input_vars') ? 'max_input_vars = 1000 (PHP default value)' :
				(ini_get('max_input_vars') < 2000 ? 'max_input_vars = ' . ini_get('max_input_vars') : ''));
			$maxInputMsg .= $maxInputMsg ? ': >= 2000 is recommended' : '';

			$phpextensions = array_map('strtolower', get_loaded_extensions());
			$phpextensionsMissing = [];
			foreach(we_dialog_sysinfo::REQUIRED_EXTENSIONS as $exten){
				if(!in_array(strtolower($exten), $phpextensions, true)){
					$phpextensionsMissing[] = $exten;
				}
			}
			$phpextensionsMissing = empty($phpextensionsMissing) ? '' : '<p>Some php extensions missing: <ul>' . implode('<li>', $phpextensionsMissing) . '</ul></p>';

			$criticalPhpMsg = trim($maxInputMsg . $phpextensionsMissing);
			if($criticalPhpMsg){
				t_e('Critical PHP Settings found', $criticalPhpMsg);
			}
		}

		if(we_base_request::string('checkLogin', '') && !$_COOKIE){
			$layout = self::getError(g_l('start', '[cookies_disabled]'), true);

			self::printLoginPage(we_users_user::LOGIN_UNKNOWN, 400, we_html_element::body(['id' => 'loginScreen'], $layout->getHtml()));
			return false;
		} else if(!we_database_base::hasDB() || DB_WE::$inst->Error === 'No database selected'){
			$layout = self::getError(g_l('start', '[no_db_connection]'));

			self::printLoginPage(we_users_user::LOGIN_UNKNOWN, 503, we_html_element::body(['id' => 'loginScreen'], $layout->getHtml()));
			return false;
		}
		return true;
	}

	private static function updateConfig(): void{
		if(!defined('CONF_SAVED_VERSION') || (defined('CONF_SAVED_VERSION') && (intval(WE_SVNREV) !== intval(CONF_SAVED_VERSION)))){
			define('WE_VERSION_UPDATE', 1);
			//resave config file(s)
			we_base_preferences::check_global_config(true);
			we_cache_file::clean('newwe_version');
		}
		we_base_file::checkAndMakeFolder(WE_DOCUMENT_ROOT . WE_THUMBNAIL_DIRECTORY);
		we_base_file::checkAndMakeFolder(WE_DOCUMENT_ROOT . WE_PREVIEW_DIR);
		we_base_preferences::writeDefaultLanguageConfig();
	}

	private static function printLoginPage(int $login, int $status = 200, string $body = '', string $msg = ''): void{
		$loginIsOK = ($login === we_users_user::LOGIN_OK);
		header('Expires: ' . gmdate('D, d.m.Y H:i:s') . ' GMT');
		header('Cache-Control: must-revalidate, post-check=0, pre-check=0');
		header('Pragma: public');
		header('Pragma: no-cache');
		we_html_tools::setHttpCode($status);

		echo we_html_tools::getHtmlTop('', we_html_element::cssLink(CSS_DIR . 'loginScreen.css') .
			($loginIsOK ?
				we_html_element::linkElement(['rel' => 'home', 'href' => WEBEDITION_DIR]) .
				we_html_element::linkElement(['rel' => 'author', 'href' => g_l('start', '[we_homepage]')]) :
				we_html_element::meta(['http-equiv' => 'refresh', 'content' => '1700; URL=/webEdition/index.php'])) .
			we_html_element::linkElement(['rel' => 'SHORTCUT ICON', 'href' => IMAGE_DIR . 'webedition.ico']) .
			we_html_element::jsScript(JS_DIR . 'weFrontendEdit_header.js') .
			we_html_element::jsScript(JS_DIR . 'we/layout/jsWindow.js') .
			we_html_element::jsScript(JS_DIR . 'login.js', '', ['id' => 'loadVarLogin', 'data-login' => setDynamicVar([
					'messageSettings' => (we_base_util::WE_MESSAGE_ERROR + we_base_util::WE_MESSAGE_WARNING + we_base_util::WE_MESSAGE_NOTICE),
					'message_reporting' => [
						'notice' => g_l('alert', '[notice]'),
						'warning' => g_l('alert', '[warning]'),
						'error' => g_l('alert', '[error]'),
					],
					'messageTypes' => [
						'WE_MESSAGE_NOTICE' => we_base_util::WE_MESSAGE_NOTICE,
						'WE_MESSAGE_WARNING' => we_base_util::WE_MESSAGE_WARNING,
						'WE_MESSAGE_ERROR' => we_base_util::WE_MESSAGE_ERROR
					],
					'loginStatus' => [
						'login' => $login,
						'msg' => $msg,
					],
					'consts' => [
						'alert' => [
							'no_cookies' => g_l('alert', '[no_cookies]'),
							'timesLoginError' => sprintf(g_l('alert', '[3timesLoginError]'), we_base_constants::LOGIN_FAILED_NR, we_base_constants::LOGIN_FAILED_TIME),
						],
						'loginTypes' => [
							'CREDENTIALS_INVALID_HOOK' => we_users_user::LOGIN_CREDENTIALS_INVALID_HOOK,
							'DENIED_TOTP' => we_users_user::LOGIN_DENIED_TOTP,
							'DENIED' => we_users_user::LOGIN_DENIED,
							'OK' => we_users_user::LOGIN_OK,
							'DENIED_SECURITY' => we_users_user::LOGIN_DENIED_SECURITY,
							'CREDENTIALS_INVALID' => we_users_user::LOGIN_CREDENTIALS_INVALID,
							'UNKNOWN' => we_users_user::LOGIN_UNKNOWN,
						],
						'dirs' => [
							'WEBEDITION_DIR' => WEBEDITION_DIR,
						],
					/* 'size' => [
					  'width' => (empty($_SESSION['prefs']['weWidth']) ? 8000 : $_SESSION['prefs']['weWidth']),
					  'height' => (empty($_SESSION['prefs']['weHeight']) ? 6000 : $_SESSION['prefs']['weHeight'])
					  ] */
					],
			])]), $body);
	}

	private static function cleanup(): void{
		/*		 * ***************************************************************************
		 * CLEAN Temporary Data left over from last logout  bug #4240
		 * *************************************************************************** */
		$removePaths = [
			WE_DOCUMENT_ROOT . '/OnlineInstaller',
			WE_DOCUMENT_ROOT . '/OnlineInstaller.php',
		];

		foreach($removePaths as $path){
			if(is_dir($path)){
				we_base_file::deleteLocalFolder($path, true);
			}
		}

		//cleanup
		we_base_file::cleanTempFiles(true);
		we_base_file::cleanWECache();
		we_navigation_cache::clean();
		we_captcha_captcha::cleanup();
		we_search_search::cleanOldEntries();
//clean Error-Log-Table
		DB_WE::$inst->query('DELETE FROM ' . ERROR_LOG_TABLE . ' WHERE `Date`<(NOW()-INTERVAL ' . we_base_constants::ERROR_LOG_HOLDTIME . ' DAY)');
		if(we_modules_info::isActive(we_modules_info::CUSTOMER)){
			DB_WE::$inst->query('DELETE FROM ' . CUSTOMER_AUTOLOGIN_TABLE . ' WHERE LastLogin_Date<(NOW()-INTERVAL ' . we_customer_customer::CUSTOMER_AUTOLOGIN_LIFETIME . ' SECOND)');
		}
		$cnt = fI('SELECT COUNT(1) FROM ' . ERROR_LOG_TABLE);

		if($cnt > we_base_constants::ERROR_LOG_MAX_ITEM_COUNT){
			DB_WE::$inst->query('DELETE FROM ' . ERROR_LOG_TABLE . ' WHERE 1 ORDER BY Date LIMIT ' . ($cnt - we_base_constants::ERROR_LOG_MAX_ITEM_THRESH));
		}
		//check if we have an old updater still blocking, remove after 30 mins
		if(is_dir(WE_DOCUMENT_ROOT . welu_update_const::WELU_DIR . 'updateClient')){
			if((filectime(WE_DOCUMENT_ROOT . welu_update_const::WELU_DIR . 'updateClient') + 30 * 60) < time()){
				we_base_file::deleteLocalFolder(WE_DOCUMENT_ROOT . welu_update_const::WELU_DIR . 'updateClient', true);
			} else {
				we_base_file::insertIntoCleanUp(welu_update_const::WELU_DIR . 'updateClient', 30 * 60);
			}
		}
		if(we_modules_info::isActive(we_modules_info::WORKFLOW)){
			we_workflow_utility::cleanup();
		}
	}

	private static function versionCheck(): void{
//updatecheck
		try{
			we_cache_file::load('newwe_version');
		} catch (we_exception_cache $e){
			try{
				$data = we_unserialize(we_http_util::getHTTP('https://update.webedition.org', '/server/we/latest.php?vers=' .
						WE_VERSION .
						'&supp=' . WE_VERSION_SUPP .
						'&branch=' . WE_VERSION_BRANCH .
						(WE_VERSION_SUPP !== 'release' ? '&beta=true' : '')));
				if(is_array($data)){
					we_cache_file::save('newwe_version', $data, we_base_util::SECONDS_WEEK, SERIALIZE_JSON);
				}
			} catch (we_exception_http $e){
				//ignore
			}
		}
	}

	public static function checkMaxLoginOK(): bool{
//CHECK FOR FAILED LOGIN ATTEMPTS
		DB_WE::$inst->query('DELETE FROM ' . FAILED_LOGINS_TABLE . ' WHERE UserTable="tblUser" AND LoginDate<(NOW() - INTERVAL ' . we_base_constants::LOGIN_FAILED_HOLDTIME . ' DAY)');

		$count = fI('SELECT COUNT(1) FROM ' . FAILED_LOGINS_TABLE . ' WHERE UserTable="tblUser" AND IP="' . DB_WE::$inst->escape($_SERVER['REMOTE_ADDR']) . '" AND LoginDate>(NOW() - INTERVAL ' . we_base_constants::LOGIN_FAILED_TIME . ' MINUTE)');

		return ($count < we_base_constants::LOGIN_FAILED_NR);
	}

	private static function getError(string $reason, bool $cookie = false): we_html_table{
		$error_count = 0;

		$error = we_html_element::b($reason) .
			/* (defined('SYSTEM_WE_SESSION') && SYSTEM_WE_SESSION ? '' :
			  (!$tmp || !(is_dir($tmp) || (is_link($tmp) && is_dir(readlink($tmp) ?: ''))) ?
			  ( ++$error_count . ' - ' . sprintf(g_l('start', '[tmp_path]'), $tmp) . we_html_element::br()) :
			  '')) . */
			(!ini_get('session.use_cookies') ?
			( ++$error_count . ' - ' . g_l('start', '[use_cookies]') . we_html_element::br()) :
			'') .
			(ini_get('session.cookie_path') != WEBEDITION_DIR ?
			( ++$error_count . ' - ' . sprintf(g_l('start', '[cookie_path]'), (string) ini_get('session.cookie_path')) . we_html_element::br()) :
			'') .
			($cookie && $error_count == 0 ?
			( ++$error_count . ' - ' . g_l('start', '[login_session_terminated]') . we_html_element::br()) :
			'') .
			we_html_element::br() . g_l('start', ($error_count == 1 ? '[solution_one]' : '[solution_more]'));

		$layout = (new we_html_table(['id' => 'errorTable'], 1, 1))
			->setCol(0, 0, we_html_tools::messageBox(500, 250, we_html_element::p([], $error), g_l('alert', '[phpError]')), ['style' => 'text-align:center;vertical-align:middle']);
		return $layout;
	}

	private static function isLoginOK(): int{
		/*		 * ***************************************************************************
		 * SWITCH MODE
		 * *************************************************************************** */
//set denied as default
		//$login = we_users_user::LOGIN_DENIED;
		if(self::$userLoginDenied){//security Token invalid or user credentials not sufficient
			return self::$userLoginDenied;
		}
		if(isset($_SESSION['user']['Username']) && isset($_POST['WE_LOGIN_password']) && isset($_POST['WE_LOGIN_username'])){
			if(($mode = we_base_request::string('mode', we_base_constants::MODE_NORMAL))){
				if(version_compare(PHP_VERSION, '7.3.0') >= 0){
					setcookie('we_mode', $mode, ['expires' => time() + 2592000, 'httponly' => true, 'samesite' => 'strict']); //	Cookie remembers the last selected mode,it will expire in one Month !
				} else {
					setcookie('we_mode', $mode, time() + 2592000); //	Cookie remembers the last selected mode, it will expire in one Month !
				}
			}
			switch(MAINTENANCE_MODE){
				case we_base_constants::MAINTENANCE_BACKEND:
				case we_base_constants::MAINTENANCE_ALL:
					return (we_base_permission::hasPerm([we_base_perm::ADMINISTRATOR]) ? we_users_user::LOGIN_OK : we_users_user::LOGIN_DENIED_MAINTENANCE);
				default:
					return we_users_user::LOGIN_OK;
			}
		}
		if(isset($_POST['WE_LOGIN_password']) && isset($_POST['WE_LOGIN_username'])){
			return we_users_user::LOGIN_CREDENTIALS_INVALID;
		}
		return we_users_user::LOGIN_UNKNOWN;
	}

	private static function getLoginOrErrorBody(): array{
		$body_javascript = '';
		$login = self::isLoginOK();
		switch($login){
			case we_users_user::LOGIN_OK:
				$httpCode = 200;

				//	Here the mode - SEEM or normal is saved in the SESSION!
				//	Perhaps this must move to another place later.
				//	Later we must check permissions as well!
				if(we_base_request::string('mode', we_base_constants::MODE_NORMAL) === we_base_constants::MODE_NORMAL){
					if(we_base_permission::hasPerm([we_base_perm::SEEM_CAN_WORK_NORMAL_MODE])){
						$_SESSION['weS']['we_mode'] = we_base_constants::MODE_NORMAL;
					} else {
						$body_javascript = g_l('SEEM', '[only_seem_mode_allowed]');
						$_SESSION['weS']['we_mode'] = we_base_constants::MODE_SEE;
					}
				} else {
					$_SESSION['weS']['we_mode'] = we_base_request::string('mode', we_base_constants::MODE_NORMAL);
				}
				if(!$body_javascript){
					//try to redirect instead of js code
					$httpCode = 303;
					header('Location: ' . WEBEDITION_DIR . 'index.php?we_cmd[0]=startWE');
				}
				return ['login' => $login, 'code' => $httpCode, 'js' => $body_javascript];
			case we_users_user::LOGIN_CREDENTIALS_INVALID_HOOK:
				$body_javascript = we_gui_login::$userLoginDeniedErr;
			case we_users_user::LOGIN_CREDENTIALS_INVALID:
				we_users_user::logLoginFailed('tblUser', we_base_request::string('WE_LOGIN_username'));
				//make it harder to guess salt/password
				usleep(1100000 + rand(0, 1000000));
				//CHECK FOR FAILED LOGIN ATTEMPTS
				$cnt = fI('SELECT COUNT(1) FROM ' . FAILED_LOGINS_TABLE . ' WHERE UserTable="tblUser" AND IP="' . DB_WE::$inst->escape($_SERVER['REMOTE_ADDR']) . '" AND LoginDate>(NOW() - INTERVAL ' . we_base_constants::LOGIN_FAILED_TIME . ' MINUTE)');

				if(empty($body_javascript) || $cnt < we_base_constants::LOGIN_FAILED_NR){
					$body_javascript = ($cnt >= we_base_constants::LOGIN_FAILED_NR ?
						sprintf(g_l('alert', '[3timesLoginError]'), we_base_constants::LOGIN_FAILED_NR, we_base_constants::LOGIN_FAILED_TIME) :
						g_l('alert', '[login_failed]'));
				}
				return ['login' => $login, 'code' => 401, 'js' => $body_javascript];
			case we_users_user::LOGIN_DENIED_SECURITY:
				return ['login' => $login, 'code' => 401, 'js' => g_l('alert', '[login_failed_security]')];
			case we_users_user::LOGIN_DENIED_TOTP:
				return ['login' => $login, 'code' => 401, 'js' => g_l('alert', '[login_failed]')];
			case we_users_user::LOGIN_DENIED_MAINTENANCE:
			case we_users_user::LOGIN_DENIED:
				return ['login' => $login, 'code' => 401, 'js' => g_l('alert', '[login_denied_for_user]')];
			default:
				return ['login' => $login, 'code' => 200, 'js' => ''];
		}
	}

	public static function startWE(): void{
		self::updateConfig();

		if(!self::checkMaxLoginOK()){
			echo we_html_tools::getHtmlTop('', we_html_element::jsScript(JS_DIR . 'weFrontendEdit_header.js') .
				we_html_element::jsScript(JS_DIR . 'we/layout/jsWindow.js') .
				we_html_element::jsElement('alert("' . sprintf(g_l('alert', '[3timesLoginError]'), we_base_constants::LOGIN_FAILED_NR, we_base_constants::LOGIN_FAILED_TIME) . '");'), we_html_element::body());
			exit();
		}

		if(!self::checkCriticalSettings()){
			//e.g. Cookie or DB not available
			exit();
		}

		['login' => $login, 'code' => $httpCode, 'js' => $msg] = self::getLoginOrErrorBody();

		/*		 * ***********************************************************************
		 * GENERATE NEEDED JAVASCRIPTS
		 * *********************************************************************** */

		$loginIsOK = ($login === we_users_user::LOGIN_OK);
		self::printLoginPage($login, $httpCode, we_html_element::body(['id' => 'loginScreen'], we_html_element::form([
					'action' => WEBEDITION_DIR . 'index.php',
					'method' => 'post',
					'name' => 'loginForm',
					'onsubmit' => 'top.checkLogin();'
					], we_html_element::hiddens([
						'checkLogin' => '1',
					]) .
					($loginIsOK ? '' : we_weTag_util::tag('formToken', ['force' => true, 'lifetime' => 1800])) .
					we_dialog_info::getDialog(true, !$loginIsOK)
				) .
				we_html_element::div(['id' => 'picCopy'], 'Copyright &copy; 4ever.eu')), $msg);



//Show page and do some housekeeping
		we_base_request::finishRequest();
		self::cleanup();
		self::versionCheck();
		//do some stuff at least each time a user logs in (or out)
		include_once(WEBEDITION_PATH . 'triggerWEtasks.php');
	}

}
