<?php
declare(strict_types=1);

/**

 *
 * LICENSE
 *
 * This source file is subject to the new BSD license that is bundled
 * with this package in the file LICENSE.txt.
 * It is also available through the world-wide-web at this URL:
 * http://framework.zend.com/license/new-bsd
 * If you did not receive a copy of the license and are unable to
 * obtain it through the world-wide-web, please send an email
 * to license@zend.com so we can send you a copy immediately.
 *
 * @category   Zend
 *
 * @copyright  Copyright (c) 2005-2015 Zend Technologies USA Inc. (http://www.zend.com)
 * @license    http://framework.zend.com/license/new-bsd     New BSD License
 * @version    $Id$
 */

/**
 * Class for sending an email.
 *
 * @category   Zend
 *
 * @copyright  Copyright (c) 2005-2015 Zend Technologies USA Inc. (http://www.zend.com)
 * @license    http://framework.zend.com/license/new-bsd     New BSD License
 */
class we_mail_znd extends we_mail_mimeMessage{
	private const PROHIBITTED_MAIL_HEADERS = ['to', 'cc', 'bcc', 'from', 'subject', 'reply-to', 'return-path', 'date', 'message-id',];
	public const headerEncoding = we_mail_mime::ENCODING_QUOTEDPRINTABLE;

	/** @var ?we_mail_TransportAbstract */
	protected static $defaultTransport = null;

	/** @var string Mail character set */
	protected $charset = 'UTF-8';

	/** @var array Mail headers	 */
	protected $headers = [];

	/** @var string From: address	 */
	protected $from = '';

	/** @var string[] To: addresses */
	protected $to = [];

	/** @var string[] Array of all recipients	 */
	protected $recipients = [];

	/** @var string Reply-To header	 */
	protected $replyTo = '';

	/** @var string Return-Path header	 */
	protected $returnPath = '';

	/** @var string Subject: header	 */
	protected $subject = '';

	/** @var string Date: header	 */
	protected $date = '';

	/** @var string Message-ID: header 	 */
	protected $messageId = '';

	/** @var ?we_mail_mimePart text/plain MIME part	 */
	protected $bodyText = null;

	/** @var ?we_mail_mimePart text/html MIME part	 */
	protected $bodyHtml = null;

	/** @var string MIME boundary string	 */
	protected $mimeBoundary = '';

	/** @var string Content type of the message */
	protected $type = '';

	/*	 * #@- */

	/**
	 * Flag: whether or not email has attachments
	 * @var boolean
	 */
	public $hasAttachments = false;

	/**
	 * Sets the default mail transport for all following uses of
	 * Mail::send();
	 *
	 * @todo Allow passing a string to indicate the transport to load
	 * @todo Allow passing in optional options for the transport to load
	 * @param  we_mail_TransportAbstract $transport
	 */
	public static function setDefaultTransport(we_mail_TransportAbstract $transport): void{
		self::$defaultTransport = $transport;
	}

	/**
	 * Gets the default mail transport for all following uses of
	 * unittests
	 *
	 * @todo Allow passing a string to indicate the transport to load
	 * @todo Allow passing in optional options for the transport to load
	 */
	public static function getDefaultTransport(): ?we_mail_TransportAbstract{
		return self::$defaultTransport;
	}

	/**
	 * Clear the default transport property
	 */
	public static function clearDefaultTransport(): void{
		self::$defaultTransport = null;
	}

	/**
	 * Set content type
	 *
	 * Should only be used for manually setting multipart content types.
	 *
	 * @param  string $type Content type
	 * @return we_mail_znd Implements fluent interface
	 * @throws we_exception_mail for types not supported by Mime
	 */
	public function setType(string $type): self{
		$allowed = [
			we_mail_mime::MULTIPART_ALTERNATIVE,
			we_mail_mime::MULTIPART_MIXED,
			we_mail_mime::MULTIPART_RELATED,
		];
		if(!in_array($type, $allowed, true)){
			/**
			 * @see Mail_Exception
			 */
			throw new we_exception_mail('Invalid content type "' . $type . '"');
		}

		$this->type = $type;
		return $this;
	}

	/**
	 * Get content type of the message
	 *
	 * @return string
	 */
	public function getType(): string{
		return $this->type;
	}

	/**
	 * Set an arbitrary mime boundary for the message
	 *
	 * If not set, Mime will generate one.
	 *
	 * @param  string    $boundary
	 * @return we_mail_znd Provides fluent interface
	 */
	public function setMimeBoundary(string $boundary): we_mail_mimeMessage{
		$this->mimeBoundary = $boundary;

		return $this;
	}

	/**
	 * Return the boundary string used for the message
	 *
	 * @return string
	 */
	public function getMimeBoundary(): string{
		return $this->mimeBoundary;
	}

	/**
	 * Sets the text body for the message.
	 *
	 * @param  string $txt
	 * @param  string $charset
	 * @param  string $encoding
	 * @return we_mail_znd Provides fluent interface
	 */
	public function setBodyText(string $txt, string $charset = '', string $encoding = we_mail_mime::ENCODING_QUOTEDPRINTABLE): self{
		if(!$charset){
			$charset = $this->charset;
		}

		$mp = new we_mail_mimePart($txt);
		$mp->encoding = $encoding;
		$mp->type = we_mail_mime::TYPE_TEXT;
		$mp->disposition = we_mail_mime::DISPOSITION_INLINE;
		$mp->charset = $charset;

		$this->bodyText = $mp;

		return $this;
	}

	/**
	 * Return text body Mime_Part or string
	 *
	 * @return ?we_mail_mimePart
	 */
	public function getBodyText(): ?we_mail_mimePart{
		return $this->bodyText;
	}

	/**
	 * Sets the HTML body for the message
	 *
	 * @param  we_mail_mimePart    $html
	 */
	protected function setBodyHtml(we_mail_mimePart $html): void{
		$this->bodyHtml = $html;
	}

	/**
	 * Return Mime_Part representing body HTML
	 *
	 * @return ?we_mail_mimePart
	 */
	public function getBodyHtml(): ?we_mail_mimePart{
		return $this->bodyHtml;
	}

	/**
	 * Adds an existing attachment to the mail message
	 *
	 * @param  we_mail_mimePart $attachment
	 * @return we_mail_znd Provides fluent interface
	 */
	public function addAttachment(we_mail_mimePart $attachment): self{
		$this->addPart($attachment);
		$this->hasAttachments = true;

		return $this;
	}

	/**
	 * Creates a Mime_Part attachment
	 *
	 * Attachment is automatically added to the mail object after creation. The
	 * attachment object is returned to allow for further manipulation.
	 *
	 * @param  string         $body
	 * @param  string         $mimeType
	 * @param  string         $disposition
	 * @param  string         $encoding
	 * @param  string         $filename OPTIONAL A filename for the attachment
	 * @return we_mail_mimePart Newly created Mime_Part object (to allow
	 * advanced settings)
	 */
	public function createAttachment(string $body, string $mimeType = we_mail_mime::TYPE_OCTETSTREAM, string $disposition = we_mail_mime::DISPOSITION_ATTACHMENT, string $encoding = we_mail_mime::ENCODING_BASE64, string $filename = ''): we_mail_mimePart{

		$mp = new we_mail_mimePart($body);
		$mp->encoding = $encoding;
		$mp->type = $mimeType;
		$mp->disposition = $disposition;
		$mp->filename = $filename;

		$this->addAttachment($mp);

		return $mp;
	}

	/**
	 * Return a count of message parts
	 *
	 * @return integer
	 */
	public function getPartCount(): int{
		return count($this->_parts);
	}

	/**
	 * Encode header fields
	 *
	 * Encodes header content according to RFC1522 if it contains non-printable
	 * characters.
	 *
	 * @param  string $value
	 * @return string
	 */
	public static function encodeHeader(string $value): string{
		return we_mail_mime::isPrintable($value) ? $value :
			/* (we_mail_znd::headerEncoding === we_mail_mime::ENCODING_QUOTEDPRINTABLE ? */
			we_mail_mime::encodeQuotedPrintableHeader($value, 'UTF-8', we_mail_mime::LINELENGTH, we_mail_mime::LINEEND) /* :
			  we_mail_mime::encodeBase64Header($value, 'UTF-8', we_mail_mime::LINELENGTH, we_mail_mime::LINEEND)
			  ) */;
	}

	/**
	 * Add a header to the message
	 *
	 * Adds a header to this message. If append is true and the header already
	 * exists, raises a flag indicating that the header should be appended.
	 *
	 * @param string  $headerName
	 * @param string  $value
	 * @param bool $append
	 */
	protected function storeHeader(string $headerName, string $value, bool $append = false): void{
		if(isset($this->headers[$headerName])){
			$this->headers[$headerName][] = $value;
		} else {
			$this->headers[$headerName] = [$value];
		}

		if($append){
			$this->headers[$headerName]['append'] = true;
		}
	}

	/**
	 * Helper function for adding a recipient and the corresponding header
	 *
	 * @param string $headerName
	 */
	protected function addRecipientAndHeader(string $headerName, we_mail_address $adr): void{
		$this->recipients[] = $adr->email;
		$this->storeHeader($headerName, $adr->format(), true);
	}

	/**
	 * Adds To-header and recipient, $email can be an array, or a single string address
	 *
	 * @return we_mail_znd Provides fluent interface
	 */
	protected function addTo(?we_mail_address $adr): self{
		if(!$adr){
			return $this;
		}
		$this->addRecipientAndHeader('To', $adr);
		$this->to[] = $adr->email;
		return $this;
	}

	/**
	 * Adds Cc-header and recipient, $email can be an array, or a single string address
	 *
	 * @return we_mail_znd Provides fluent interface
	 */
	public function addCc(?we_mail_address $adr): self{
		if(!$adr){
			return $this;
		}
		$this->addRecipientAndHeader('Cc', $adr);
		return $this;
	}

	/**
	 * Adds Bcc recipient, $email can be an array, or a single string address
	 *
	 * @return we_mail_znd Provides fluent interface
	 */
	public function addBcc(?we_mail_address $adr): self{
		if(!$adr){
			return $this;
		}
		$this->addRecipientAndHeader('Bcc', $adr);
		return $this;
	}

	/**
	 * Return list of recipient email addresses
	 *
	 * @return string[]
	 */
	public function getRecipients(): array{
		return $this->recipients;
	}

	/**
	 * Clear header from the message
	 *
	 * @param string $headerName
	 * @return we_mail_znd Provides fluent inter
	 */
	public function clearHeader(string $headerName): self{
		if(isset($this->headers[$headerName])){
			unset($this->headers[$headerName]);
		}
		return $this;
	}

	/**
	 * Clears list of recipient email addresses
	 *
	 * @return we_mail_znd Provides fluent interface
	 */
	public function clearRecipients(): self{
		$this->recipients = [];
		$this->to = [];

		$this->clearHeader('To');
		$this->clearHeader('Cc');
		$this->clearHeader('Bcc');

		return $this;
	}

	/**
	 * Sets From-header and sender of the message
	 *
	 * @return we_mail_znd Provides fluent interface
	 * @throws we_exception_mail if called subsequent times
	 */
	public function setFrom(?we_mail_address $adr): self{
		if(!$adr){
			return $this;
		}
		if($this->from){
			/**
			 * @see Mail_Exception
			 */
			throw new we_exception_mail('From Header set twice');
		}

		$this->from = $adr->email;
		$this->storeHeader('From', $adr->format(), true);

		return $this;
	}

	/**
	 * Set Reply-To Header
	 *
	 * @return we_mail_znd
	 * @throws we_exception_mail if called more than one time
	 */
	public function setReplyTo(?we_mail_address $adr): self{
		if(!$adr){
			return $this;
		}
		if($this->replyTo !== ''){
			throw new we_exception_mail('Reply-To Header set twice');
		}

		$this->replyTo = $adr->email;
		$this->storeHeader('Reply-To', $adr->format(), true);

		return $this;
	}

	/**
	 * Returns the sender of the mail
	 *
	 * @return string
	 */
	public function getFrom(): string{
		return $this->from;
	}

	/**
	 * Returns the current Reply-To address of the message
	 *
	 * @return string Reply-To address, null when not set
	 */
	public function getReplyTo(): string{
		return $this->replyTo;
	}

	/**
	 * Clears the sender from the mail
	 *
	 * @return we_mail_znd Provides fluent interface
	 */
	public function clearFrom(): we_mail_mimeMessage{
		$this->from = '';
		$this->clearHeader('From');

		return $this;
	}

	/**
	 * Clears the current Reply-To address from the message
	 *
	 * @return we_mail_znd Provides fluent interface
	 */
	public function clearReplyTo(): we_mail_mimeMessage{
		$this->replyTo = '';
		$this->clearHeader('Reply-To');

		return $this;
	}

	/**
	 * Sets the Return-Path header of the message
	 *
	 * @return we_mail_znd Provides fluent interface
	 * @throws we_exception_mail if set multiple times
	 */
	public function setReturnPath(we_mail_address $adr): self{
		if($this->returnPath !== ''){
			throw new we_exception_mail('Return-Path Header set twice');
		}
		$this->returnPath = $adr->email;
		$this->storeHeader('Return-Path', $adr->format(), false);

		return $this;
	}

	/**
	 * Returns the current Return-Path address of the message
	 *
	 * If no Return-Path header is set, returns the value of {@link $_from}.
	 *
	 * @return string
	 */
	public function getReturnPath(): string{
		if($this->returnPath !== ''){
			return $this->returnPath;
		}

		return $this->from;
	}

	/**
	 * Clears the current Return-Path address from the message
	 *
	 * @return we_mail_znd Provides fluent interface
	 */
	public function clearReturnPath(): self{
		$this->returnPath = '';
		$this->clearHeader('Return-Path');

		return $this;
	}

	/**
	 * Sets the subject of the message
	 *
	 * @param   string    $subject
	 * @return  we_mail_znd Provides fluent interface
	 * @throws  we_exception_mail
	 */
	public function setSubject(string $subject): we_mail_mimeMessage{
		if($this->subject !== ''){
			throw new we_exception_mail('Subject set twice');
		}
		$subject = $this->filterOther($subject);
		$this->subject = self::encodeHeader($subject);
		$this->storeHeader('Subject', $this->subject);

		return $this;
	}

	/**
	 * Returns the encoded subject of the message
	 *
	 * @return string
	 */
	public function getSubject(): string{
		return $this->subject;
	}

	/**
	 * Clears the encoded subject from the message
	 *
	 * @return  we_mail_znd Provides fluent interface
	 */
	public function clearSubject(): self{
		$this->subject = '';
		$this->clearHeader('Subject');

		return $this;
	}

	/**
	 * Sets Date-header
	 *
	 * @param  int|string|DateTime $date
	 * @return we_mail_mimeMessage Provides fluent interface
	 * @throws we_exception_mail if called subsequent times or wrong date format.
	 */
	public function setDate($date = null): we_mail_mimeMessage{
		if($this->date !== ''){
			throw new we_exception_mail('Date Header set twice');
		}
		if($date === null){
			$date = date('r');
		} else if(is_int($date)){
			$date = date('r', $date);
		} else if(is_string($date)){
			$date = strtotime($date);
			if($date === false || $date < 0){
				/**
				 * @see Mail_Exception
				 */
				throw new we_exception_mail('String representations of Date Header must be strtotime()-compatible');
			}
			$date = date('r', $date);
		} else {
			/**
			 * @see Mail_Exception
			 */
			throw new we_exception_mail(__METHOD__ . ' only accepts UNIX timestamps and strtotime()-compatible strings');
		}
		$this->date = $date;
		$this->storeHeader('Date', $date);

		return $this;
	}

	/**
	 * Returns the formatted date of the message
	 *
	 * @return string
	 */
	public function getDate(): string{
		return $this->date;
	}

	/**
	 * Clears the formatted date from the message
	 *
	 * @return we_mail_znd Provides fluent interface
	 */
	public function clearDate(): we_mail_mimeMessage{
		$this->date = '';
		$this->clearHeader('Date');

		return $this;
	}

	/**
	 * Sets the Message-ID of the message
	 *
	 * @param   string  $id
	 * true  :Auto
	 * false :No set
	 * null  :No set
	 * string:Sets given string (Angle brackets is not necessary)
	 * @return  we_mail_znd Provides fluent interface
	 * @throws  we_exception_mail
	 */
	public function setMessageId(string $id = ''): self{
		if($this->messageId !== ''){
			throw new we_exception_mail('Message-ID set twice');
		}
		$id = $this->filterOther($id);
		$this->messageId = $id;
		$this->storeHeader('Message-Id', '<' . $this->messageId . '>');

		return $this;
	}

	/**
	 * Returns the Message-ID of the message
	 *
	 * @return string
	 */
	public function getMessageId(): string{
		return $this->messageId;
	}

	/**
	 * Clears the Message-ID from the message
	 *
	 * @return we_mail_znd Provides fluent interface
	 */
	public function clearMessageId(): we_mail_mimeMessage{
		$this->messageId = '';
		$this->clearHeader('Message-Id');

		return $this;
	}

	/**
	 * Creates the Message-ID
	 *
	 * @return string
	 */
	public function createMessageId(): string{
		$time = time();

		$user = ($this->from ?: ($_SERVER['REMOTE_ADDR'] ?? getmypid()) );

		$rand = mt_rand();

		$recipient = ($this->recipients ?
			reset($this->recipients) :
			'unknown');

		$hostName = ( $_SERVER['SERVER_NAME'] ?? php_uname('n'));


		return sha1($time . $user . $rand . $recipient) . '@' . $hostName;
	}

	/**
	 * Add a custom header to the message
	 *
	 * @param  string              $name
	 * @param  string              $value
	 * @param  boolean             $append
	 * @return we_mail_znd           Provides fluent interface
	 * @throws we_exception_mail on attempts to create standard headers
	 */
	public function addHeader(string $name, string $value, bool $append = false): self{
		if(in_array(strtolower($name), self::PROHIBITTED_MAIL_HEADERS, true)){
			throw new we_exception_mail('Cannot set standard header from addHeader()');
		}

		$this->storeHeader($name, self::encodeHeader($this->filterOther($value)), $append);

		return $this;
	}

	/**
	 * Return mail headers
	 *
	 * @return array
	 */
	public function getHeaders(): array{
		return $this->headers;
	}

	/**
	 * Sends this email using the given transport or a previously
	 * set DefaultTransport or the internal mail function if no
	 * default transport had been set.
	 *
	 * @return bool
	 */
	public function send(): bool{
		$transport = (!self::$defaultTransport instanceof we_mail_TransportAbstract ?
			new we_mail_TransportSendmail() :
			self::$defaultTransport);

		if($this->date === ''){
			$this->setDate();
		}

		$transport->send($this);

		return true;
	}

	/**
	 * Filter of other data
	 *
	 * @param string $data
	 * @return string
	 */
	protected function filterOther(string $data): string{
		return strtr($data, [
			"\r" => '',
			"\n" => '',
			"\t" => '',
		]);
	}

}
