/* global WE, top,we_cmd */

/**
 * webEdition CMS
 *
 * $Rev: 6329b796be62 $
 * $Author: Lukas Imhof $
 * $Date: 2020-11-10 16:09 +0100 $
 *
 * This source is part of webEdition CMS. webEdition CMS is
 * free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * any later version.
 *
 * The GNU General Public License can be found at
 * http://www.gnu.org/copyleft/gpl.html.
 * A copy is found in the textfile
 * webEdition/licenses/webEditionCMS/License.txt
 *
 * @license	http://www.gnu.org/copyleft/gpl.html  GPL
 */
'use strict';

/**
 // CLASS:
 //   EditorFrameController
 //
 // DESCRIPTION:
 //   manages all available EditorFrames
 //
 // ARGUMENTS:
 //   none
 //
 // RETURNS:
 //   none
 */
WE().layout.EditorFrameController = function (startCmd) {
	// reference to the Editor Frameset: used to manipulate the "cols" attribute
	this.MultiEditorFrameset = null;
	// reference to the window containing the EditorFrameset, check if we need this
	this.MultiEditorFramesetWindow = null;
	// currently active Frameset, id of the frame
	this.ActiveEditorFrameId = null;
	// array/object of the Editor Frames (EditorFrame)
	this.EditorFrames = null;
	// array of ids of free EditorFramesets
	this.FreeEditorFrames = null;
	// amount of editorFrames
	this.EditorWindowsAmount = 0;
	this.startCmd = startCmd;
};

WE().layout.EditorFrameController.prototype = {
	/*functions concerning the functionality*/
	goHome: function () {
		if (this.startCmd) {
			switch (this.startCmd[0]) {
				case 'toggleFrames':
					this.toggleFrames();
					break;
				case 'blankEditor':
					var nextWindow;
					if (!this.getNumberOfFreeWindows() && (nextWindow = this.getFreeWindow())) {
						WE().layout.multiTabs.addTab(nextWindow.getFrameId());
						this.closeDocument(nextWindow.getFrameId());
					}
					break;
				default:
					this.openDocument.apply(this, this.startCmd);
			}
		}
	},
	/**
	 // FUNCTION:
	 //   init
	 //
	 // DESCRIPTION:
	 //   initailises the EditorFrameControler
	 //
	 // ARGUMENTS:
	 //   none
	 //
	 // RETURNS:
	 //   nothing
	 */
	init: function () {
		// init references to FramesetWindow and Frameset
		this.MultiEditorFramesetWindow = top;
		this.MultiEditorFrameset = this.MultiEditorFramesetWindow.document.getElementById("multiEditorFrameset");

		var _frames = this.MultiEditorFramesetWindow.document.getElementsByTagName("iframe");

		if (!_frames.length) {
			return;
		}
		this.EditorFrames = {};
		this.FreeEditorFrames = [];
		this.EditorWindowsAmount = _frames.length;

		for (var i = 0; i < _frames.length; i++) {
			if (!_frames[i].id.match(/^multiEditFrame_/)) {
				continue;
			}
			this.EditorFrames[_frames[i].id] = new (WE().layout.EditorFrame)(_frames[i], _frames[i].id);
			this.FreeEditorFrames.push(_frames[i].id);

		}
	},
	/**FUNCTION:
	 //   isInitialized
	 //
	 // DESCRIPTION:
	 //   returns if the EditorFrameControlle is initialized
	 //
	 // ARGUMENTS:
	 //   none
	 //
	 // RETURNS:
	 //   boolean
	 --------------------------------------------------------------------*/
	isInitialized: function () {
		return (this.EditorFrames !== null);
	},
	/**
	 * FUNCTION:
	 *	openDocument
	 *
	 * DESCRIPTION:
	 *	tries to open a document in a (new) editor window.
	 *	in Expert mode:
	 *		Switch Editor to the document if it is already open | open document in a new editor if possible
	 *	in SeeMode:
	 *		Open document in window, ask to save unsaved changes
	 *
	 *	possible we_cmds are:
	 *		- new_document
	 *		- edit_document
	 *
	 * ARGUMENTS:
	 *   table: table of the document
	 *   id: id of the document
	 *   ct: contenttype of the document
	 *   editcmd: other command like cockpit, etc.
	 *   dt: ??? (transaction??)
	 *   url: url when opening a document via URL (SeeMode)
	 *   code: ?? (initial code for the document)
	 *   mode: ??
	 *
	 * RETURNS:
	 *   nothing
	 */
	openDocument: function (table, id, ct, editcmd, dt, url, code, mode, parameters) {
		if (this.EditorFrames === null) {
			this.init();
		}
		var editorId;

		// initalize variables
		dt = dt ? dt : ""; // doctype if open a document via doctype!
		editcmd = editcmd ? editcmd : ""; // editcmd like open_cockpit
		url = url ? url : ""; // doctype if open a document via doctype!
		code = code ? code : ""; // doctype if open a document via doctype!
		mode = mode ? mode : ""; // doctype if open a document via doctype!
		parameters = parameters ? parameters : "";

		// editcmd can be one of open_cockpit

		// check if a already open document shall be opened
		if ((editorId = this.getEditorIdOfOpenDocument(table, id, editcmd, url))) { // activate open tab
			if (parameters !== this.getEditorFrame(editorId).getEditorDocumentParameters()) {
				// re-open document
				this.closeDocument(editorId, ['openDocumentAllParams', table, id, ct, editcmd, dt, url, code, mode, parameters]);
			} else if (this.ActiveEditorFrameId !== editorId) {
				// add to navigationHistory:
				WE().layout.navigationHistory.addDocToHistory(table, id, ct, editcmd, url, parameters);

				// activate in tree
				if (top.treeData) {
					if (id) {
						top.treeData.selectNode(table, id);
					} else {
						top.treeData.unselectNode();
					}
				}

				// activate tab
				WE().layout.multiTabs.openTab(editorId);
				this.setActiveEditorFrame(editorId);
				this.toggleFrames();
			}
			return;
		}
		// open new frame if possible

		if (this.getNumberOfFreeWindows() > 0) { // if there is a free frame, use it
			if (editcmd || url || id) {	// add to navigationHistory
				WE().layout.navigationHistory.addDocToHistory(table, id, ct, editcmd, url, parameters);
			}

			if (editcmd) { // open cockpit !
				// deactivate tree
				if (top.treeData) {
					top.treeData.unselectNode();
				}
				top.we_cmd(editcmd);

			} else if (url) {
				we_cmd('open_extern_document', url, parameters);
			} else if (id) { // edit_document
				if (parameters) {
					we_cmd('edit_document_with_parameters', table, id, (ct ? ct : ''), parameters);
				} else {
					// instead calling the command we could also build the url and call it from here
					we_cmd('edit_document', table, id, (ct ? ct : ''));
				}
			} else if (code !== undefined && code) { // open new document with standard code
				we_cmd('new_document', table, id, ct, "", "", "", "", dt, "", code);
			} else {
				we_cmd('new_document', table, id, ct, "", "", "", "", dt);
			}
			return;
		}
		if (this.EditorWindowsAmount === 1) { // only one active document here, for example SeeMode
			// build nextCmd
			// table,id,ct,editcmd,dt,url,code,mode
			this.closeDocument(this.ActiveEditorFrameId, ['openDocumentAllParams', table, id, ct, editcmd, dt, url, code, mode, parameters]);
			return;
		}
		WE().util.showMessage(WE().consts.g_l.main.no_editor_left, WE().consts.message.WE_MESSAGE_ERROR, window);
	},
	/**
	 * FUNCTION:
	 *   closeDocument
	 *
	 * DESCRIPTION:
	 *	This function is called when a document of the editor is closed
	 *	Here are checks, if the document has unsaved changes
	 *
	 * ARGUMENTS:
	 *   frameId - id of the editor window, which should be freed
	 *   nextCommand - string
	 *
	 * RETURNS:
	 *   nothing
	 */
	closeDocument: function (editorId, nextCommand) {
		if (this.EditorFrames === null) {
			this.init();
		}

		nextCommand = (nextCommand ? nextCommand : "");

		if (!top.we_cmd("eplugin_exit_doc", this.getEditorFrame(editorId).getEditorTransaction())) {
			return;
		}
		if (!this.EditorFrames[editorId]) {
			WE().layout.multiTabs.closeTab(editorId);
			return;
		}
		// check if there are unsaved changes
		if (this.getEditorFrame(editorId).getEditorIsHot()) {
			this.showEditor(editorId);
			top.we_cmd("exit_doc_question", editorId, this.getEditorFrame(editorId).getEditorEditorTable(), nextCommand);
			return;
		}
		// free frame select next active frame
		this.closeEditorFrame(editorId);
		WE().layout.multiTabs.closeTab(editorId);

		if (nextCommand) {
			top.we_cmd.apply(window, nextCommand);
		}
	},
	closeDocuments: function (table, deletedIds) {
		var
			usedEditors = this.getEditorsInUse(),
			openEditors = [];
		for (var frameId in usedEditors) {
			if (table === usedEditors[frameId].getEditorEditorTable() &&
				deletedIds.indexOf(usedEditors[frameId].getEditorDocumentId()) !== -1
				) {
				openEditors.push(usedEditors[frameId]);
			}
		}
		this.closeAllEditors(openEditors, true);
	},
	/**
	 * FUNCTION:
	 *   doLogout
	 *
	 * DESCRIPTION:
	 *	On logout all editors must be checked for unsaved changes
	 *
	 * ARGUMENTS:
	 *   none
	 *
	 * RETURNS:
	 *   boolean
	 */
	doLogoutMultiEditor: function () {
		// close all none Hot Editors
		if (this.FreeEditorFrames.length !== this.EditorWindowsAmount) {
			var _UsedEditors = this.getEditorsInUse();
			for (var frameId in _UsedEditors) {
				// remove all from editor-plugin
				top.we_cmd("remove_from_editor_plugin", _UsedEditors[frameId].getEditorTransaction());
				if (!_UsedEditors[frameId].getEditorIsHot()) {
					this.closeDocument(frameId);
				}
			}
		}

		// if all Editors are closed,
		if (this.FreeEditorFrames.length === this.EditorWindowsAmount ||
			this.getAllHotEditors().length === 0) {
			return true;
		}
		if ((this.EditorWindowsAmount - this.FreeEditorFrames.length) === 1) { // seeMode
			this.closeDocument(this.ActiveEditorFrameId, 'dologout');
		} else {
			top.we_cmd("exit_multi_doc_question", 'dologout');
		}
		return false;
	},
	getAllHotEditors: function () {
		var allHotDocuments = this.getEditorsInUse();
		var hot = [];
		for (var frameId in allHotDocuments) {
			if (allHotDocuments[frameId].getEditorIsHot()) {
				hot.push(allHotDocuments[frameId]);
			}
		}
		return hot;
	},
	/**
	 * FUNCTION:
	 *   closeAllDocuments
	 *
	 * DESCRIPTION:
	 *	If all editors are closed, all editors must be checked for unsaved changes
	 *
	 * ARGUMENTS:
	 *   none
	 *
	 * RETURNS:
	 *   nothing
	 */
	closeAllDocuments: function () {
		if (top.we_cmd("editor_plugin_doc_count") === 0 || window.confirm(WE().consts.g_l.main.eplugin_exit_doc)) {
			// close all none Hot Editors
			if (this.FreeEditorFrames.length !== this.EditorWindowsAmount) {
				var _UsedEditors = this.getEditorsInUse();
				for (var frameId in _UsedEditors) {
					// remove from editor plugin
					top.we_cmd("remove_from_editor_plugin", _UsedEditors[frameId].getEditorTransaction());
					if (!_UsedEditors[frameId].getEditorIsHot()) {
						this.closeDocument(frameId);
					}
				}
			}

			// if all Editors are closed,
			if (this.FreeEditorFrames.length === this.EditorWindowsAmount ||
				this.getAllHotEditors().length === 0) {
				return true;
			}
			if ((this.EditorWindowsAmount - this.FreeEditorFrames.length) === 1) { // only one document open
				this.closeDocument(this.ActiveEditorFrameId, ['close_all_documents']);
			} else {
				top.we_cmd("exit_multi_doc_question", 'close_all_documents');
			}

			return false;
		}
	},
	closeAllEditors: function (editors, force) {
		for (var i = 0; i < editors.length; i++) {
			if (force) {
				editors[i].setEditorIsHot(false);
			}
			this.closeDocument(editors[i].getFrameId());
		}
	},
	closeAllButActiveDocument: function (activeId) {
		if (top.we_cmd("editor_plugin_doc_count") === 0 || window.confirm(WE().consts.g_l.main.eplugin_exit_doc)) {
			// only do something, if more than one editor is open
			if ((this.EditorWindowsAmount - this.FreeEditorFrames.length) > 1) {
				// get active id, if not given
				if (!activeId) {
					activeId = this.ActiveEditorFrameId;
				}

				var _UsedEditors = this.getEditorsInUse();
				var frameId;
				// remove all from editor plugin
				for (frameId in _UsedEditors) {
					if (frameId !== activeId) {
						top.we_cmd("remove_from_editor_plugin", _UsedEditors[frameId].getEditorTransaction());
					}
				}

				_UsedEditors = this.getEditorsInUse();
				// close all none Hot editors
				for (frameId in _UsedEditors) {
					if (frameId !== activeId) {
						if (_UsedEditors[frameId].getEditorIsHot()) {
							this.closeDocument(frameId, ['close_all_but_active_document', activeId]);
							return;

						}
						this.closeDocument(frameId);
					}
				}
			}
		}
	},
	/**-------------------------------------------------------------------
	 // FUNCTION:
	 //   closeEditorFrame
	 //
	 // DESCRIPTION:
	 //   Clears the current editor and frees it for the next document
	 //
	 // ARGUMENTS:
	 //   frameId - id of the editor window, which should be freed
	 //
	 // RETURNS:
	 //   nothing
	 --------------------------------------------------------------------*/
	closeEditorFrame: function (frameId) {
		var docRef;
		if (this.EditorFrames[frameId]) {
			switch (this.EditorFrames[frameId].EditorType) {
				case "cockpit":
					docRef = this.EditorFrames[frameId].getDocumentReference();
					// close all modal dialogs
					if (docRef.closeAllModalWindows) {
						docRef.closeAllModalWindows();
						if (docRef.isHot()) {
							// save changes, in cockpit
							docRef.saveSettings();
						}
					}
					break;
				case "model":
					docRef = this.EditorFrames[frameId].getDocumentReference();
					if (docRef.closeAllModalWindows) {
						docRef.closeAllModalWindows();
					}
					// unlock document
					var trans = this.EditorFrames[frameId].getEditorTransaction();
					if (trans) {
						top.we_cmd('we_users_unlock', this.EditorFrames[frameId].getEditorDocumentId(), WE().session.userID, this.EditorFrames[frameId].getEditorEditorTable(), trans);
						top.we_cmd("remove_from_editor_plugin", trans);
					}

					break;
			}

			// remove from tree, if possible
			// deactivate in tree
			if (top.treeData && top.treeData.table === this.getEditorFrame(frameId).getEditorEditorTable() && this.ActiveEditorFrameId === frameId) {
				top.treeData.unselectNode();
			}

			// about:blank
			this.EditorFrames[frameId].freeEditor();

			// add to free frames
			this.FreeEditorFrames.push(frameId);

			// make other frame active, if the closed one was active
			if (this.ActiveEditorFrameId === frameId) { // active frame was closed, show next.

				this.ActiveEditorFrameId = null;
				var _tmpKey = null;
				if (this.FreeEditorFrames.length !== this.EditorWindowsAmount) { // there are filled frames left

					var _reachedCurrent = false;
					this.ActiveEditorFrameId = null;
					for (var frameKey in this.EditorFrames) {

						if (!_reachedCurrent || _tmpKey === null) {
							if (this.EditorFrames[frameKey].getEditorIsInUse()) {
								_tmpKey = frameKey;

							}
							if (frameKey === frameId) {
								_reachedCurrent = true;

							}
						}

					}
					this.showEditor(_tmpKey);
					this.ActiveEditorFrameId = _tmpKey;
				}
			}
			this.toggleFrames();
		}
	},
	/*FUNCTION:
	 //   showEditor
	 //
	 // DESCRIPTION:
	 //   Activates the EditorFrame with the given Id
	 //
	 // ARGUMENTS:
	 //   editorId - id of the editor window
	 //
	 // RETURNS:
	 //   nothing
	 */
	showEditor: function (editorId) {
		if (editorId !== this.ActiveEditorFrameId) {
			// add to navigationHistory:
			var _currentEditor = this.getEditorFrame(editorId);

			if (_currentEditor.getEditorIsInUse()) {

				WE().layout.navigationHistory.addDocToHistory(
					_currentEditor.getEditorEditorTable(),
					_currentEditor.getEditorDocumentId(),
					_currentEditor.getEditorContentType(),
					_currentEditor.getEditorEditCmd(),
					_currentEditor.getEditorUrl(),
					_currentEditor.getEditorDocumentParameters()
					);

				// activate tab
				WE().layout.multiTabs.openTab(editorId);

				// highlight tree
				if (top.treeData) {
					if (parseInt(_currentEditor.getEditorDocumentId())) {
						top.treeData.selectNode(_currentEditor.getEditorEditorTable(), _currentEditor.getEditorDocumentId());
					} else {
						top.treeData.unselectNode();

					}
				}
				this.setActiveEditorFrame(editorId);
				this.toggleFrames();
			}
		}
	},
	switchToContentEditor: function (toContentEditor) {
		this.getActiveEditorFrame().switchToContentEditor(toContentEditor);
	},
	getVisibleEditorFrame: function () {
		var editorFrame = this.getActiveEditorFrame();
		if (!editorFrame) {
			return null;
		}
		return editorFrame.getContentEditor();
	},
	isEditTab: function () {
		var editorFrame = this.getActiveEditorFrame();
		if (!editorFrame) {
			return null;
		}
		return editorFrame.getContentEditor() === this.getActiveDocumentReference().frames[2];
	},
	/**FUNCTION:
	 //   toggleFrames
	 //
	 // DESCRIPTION:
	 //   sets the "cols"-Attribute of the EditorFrameset
	 //
	 // ARGUMENTS:
	 //   none
	 //
	 // RETURNS:
	 //   nothing
	 */
	toggleFrames: function () {
		var frameId,
			isChrome = WE().session.isChrome;
		//		var _colStr = "";
		if (!this.ActiveEditorFrameId) {
			var first = true;
			for (frameId in this.EditorFrames) {
				if (first) {
					this.getEditorFrame(frameId).setEmptyEditor();
					if (WE().session.isChrome) {
						this.getEditorFrame(frameId).EditorFrameReference.style.display = "block";
					} else {
						this.getEditorFrame(frameId).EditorFrameReference.style.width = "100%";
						this.getEditorFrame(frameId).EditorFrameReference.style.height = "100%";
					}
					first = false;
				} else {
					if (WE().session.isChrome) {
						this.getEditorFrame(frameId).EditorFrameReference.style.display = "none";
					} else {
						this.getEditorFrame(frameId).EditorFrameReference.style.height = "0";
						this.getEditorFrame(frameId).EditorFrameReference.style.witdh = "0";
					}
				}
			}
			return;
		}

		for (frameId in this.EditorFrames) {
			if (this.ActiveEditorFrameId === frameId) {
				if (isChrome) {
					this.getEditorFrame(frameId).EditorFrameReference.style.display = "block";
				} else {
					this.getEditorFrame(frameId).EditorFrameReference.style.width = "100%";
					this.getEditorFrame(frameId).EditorFrameReference.style.height = "100%";
				}
			} else {
				if (this.getEditorFrame(frameId).getEditorIsInUse() && this.getEditorFrame(frameId).EditorType !== "none_webedition" && this.EditorFrames[frameId].getDocumentReference().closeAllModalWindows) {
					this.EditorFrames[frameId].getDocumentReference().closeAllModalWindows();
				}
				if (isChrome) {
					this.getEditorFrame(frameId).EditorFrameReference.style.display = "none";
				} else {
					this.getEditorFrame(frameId).EditorFrameReference.style.height = "0";
					this.getEditorFrame(frameId).EditorFrameReference.style.witdh = "0";
				}
			}
		}
	},
	/*FUNCTION:
	 //   getFreeWindow
	 //
	 // DESCRIPTION:
	 //   returns next free frame
	 //
	 // ARGUMENTS:
	 //   none
	 //
	 // RETURNS:
	 //   name EditorFrame object or false
	 */
	getFreeWindow: function () {
		if (this.EditorFrames === null) {
			this.init();
		}

		if (this.FreeEditorFrames.length > 0) {
			return this.EditorFrames[this.FreeEditorFrames.shift()];
			//var objref = this.EditorFrames[this.FreeEditorFrames.shift()];
			//objref.useEditorFrame(); - if there occure any errors use this here
			//return objref;
		}
		return false;
	},
	/*
	 // FUNCTION:
	 //   getNumberOfFreeWindows
	 //
	 // DESCRIPTION:
	 //   returns number of available free frames
	 //
	 // ARGUMENTS:
	 //   none
	 //
	 // RETURNS:
	 //   integer
	 */
	getNumberOfFreeWindows: function () {
		return (this.FreeEditorFrames === null ? 0 : this.FreeEditorFrames.length);
	},
	getEditorIdOfOpenDocument: function (table, id, editcmd, url) {
		if (id === "0" && !editcmd && !url) {
			return null;
		}

		for (var _editorId in this.EditorFrames) {
			if (table && id && this.getEditorEditorTable(_editorId) === table && this.getEditorDocumentId(_editorId) == id) { // open by id
				return _editorId;
			}
			if (editcmd === "open_cockpit" && this.EditorFrames[_editorId].EditorType === "cockpit") { // open a cmd window
				return _editorId;
			}
			if (url && this.EditorFrames[_editorId].EditorUrl === url) { // open with URL
				return _editorId;
			}
		}
		return null;
	},
	getEditorFrameByExactParams: function (id, table, tab, transaction) {
		if (!id || !table || !tab || !transaction) {
			return null;
		}

		var usedEditors = this.getEditorsInUse(),
			frameId, editor;

		for (frameId in this.getEditorsInUse()) {
			editor = usedEditors[frameId];
			if (editor.getEditorEditorTable() === table && parseInt(editor.getEditorDocumentId()) === parseInt(id) &&
				parseInt(editor.getEditorEditPageNr()) === parseInt(tab) && editor.getEditorTransaction() === transaction) {

				return this.getEditorFrame(frameId);
			}
		}

		return null;
	},
	getEditorTransactionByIdTable: function (id, table) {
		if (!id || !table) {
			return '';
		}

		var usedEditors = this.getEditorsInUse(),
			frameId, editor;

		for (frameId in this.getEditorsInUse()) {
			editor = usedEditors[frameId];
			if (editor.getEditorEditorTable() === table && parseInt(editor.getEditorDocumentId()) === id) {

				return this.getEditorTransaction(frameId);
			}
		}

		return '';
	},
	getEditorsInUse: function () {
		var _ret = {};
		for (var frameId in this.EditorFrames) {
			if (this.EditorFrames[frameId].getEditorIsInUse()) {
				_ret[frameId] = this.EditorFrames[frameId];
			}
		}
		return _ret;
	},
	//----------------------------------------
	// getters
	// all getters can have the transactionnumber as parameter,
	// but take the current editor as default
	//----------------------------------------

	getActiveEditorFrame: function () {
		return (this.ActiveEditorFrameId ? this.EditorFrames[this.ActiveEditorFrameId] : false);
	},
	getEditorFrameByTransaction: function (theTransaction) {
		for (var frameId in this.EditorFrames) {
			if (this.EditorFrames[frameId] && (this.EditorFrames[frameId].getEditorTransaction() === theTransaction)) {
				return this.EditorFrames[frameId];
			}
		}
		return null;
	},
	getEditorFrame: function (frameId) {
		if (frameId !== undefined && frameId !== "") {
			return this.EditorFrames[frameId];
		}
		if (this.ActiveEditorFrameId) {
			return this.EditorFrames[this.ActiveEditorFrameId];
		}
		return false;
	},
	getDocumentReferenceByTransaction: function (theTransaction) {
		var _win = this.getEditorFrameByTransaction(theTransaction);
		return (_win ? _win.getDocumentReference() : false);

	},
	getActiveDocumentReference: function () {
		if (this.EditorFrames === null) {
			this.init();
		}

		var _win = this.getActiveEditorFrame();
		return (_win ? _win.getDocumentReference() : false);
	},
	//----------------------------------------
	// setters
	//----------------------------------------
	setActiveEditorFrame: function (id) {
		if (this.ActiveEditorFrameId !== id) {
			var _oldActive;
			if ((_oldActive = this.getEditorFrame(this.ActiveEditorFrameId))) {
				_oldActive.setEditorIsActive(false);
			}

			this.ActiveEditorFrameId = id;

			var _EditorWindow = this.getEditorFrame(id);
			_EditorWindow.setEditorIsActive(true);
		}
	},
	setEditorIsHot: function (newVal, id) {
		var _EditorWindow = this.getEditorFrame(id);
		if (_EditorWindow) {
			_EditorWindow.setEditorIsHot(newVal);
		}
	},
	//----------------------------------------
	// getters
	//----------------------------------------

	getEditorIsHot: function (id) {
		var _EditorWindow = this.getEditorFrame(id);
		if (_EditorWindow) {
			return _EditorWindow.getEditorIsHot();
		}
		return null;
	},
	getEditorTransaction: function (frameId) {
		var _EditorFrame = this.getEditorFrame(frameId);
		if (_EditorFrame) {
			return _EditorFrame.getEditorTransaction();
		}
		return null;

	},
	getEditorDocumentId: function (frameId) {
		var _EditorFrame = this.getEditorFrame(frameId);
		if (_EditorFrame) {
			return _EditorFrame.getEditorDocumentId();
		}
		return null;
	},
	getEditorEditorTable: function (frameId) {
		var _EditorFrame = this.getEditorFrame(frameId);
		return _EditorFrame.getEditorEditorTable();
	},
	getEditorEditPageNr: function (frameId) {
		var _EditorFrame = this.getEditorFrame(frameId);
		return _EditorFrame.getEditorEditPageNr();
	},
	getEditorIsActive: function (frameId) {
		var _EditorFrame = this.getEditorFrame(frameId);
		return _EditorFrame.getEditorIsActive(frameId);
	},
	getEditorIsInUse: function (frameId) {
		var _EditorFrame = this.getEditorFrame(frameId);
		return _EditorFrame.getEditorIsInUse();
	},
	doReloadCmd: function (args, url, hot) {
		if (hot) {
			// set Editor hot
			var _EditorFrame = WE().layout.weEditorFrameController.getActiveEditorFrame();
			_EditorFrame.setEditorIsHot(true);
		}
		if (top.setScrollTo) {
			window.setScrollTo();
		}
		// get editor root frame of active tab
		var _currentEditorRootFrame = WE().layout.weEditorFrameController.getActiveDocumentReference();
		// get visible frame for displaying editor page
		var _visibleEditorFrame = WE().layout.weEditorFrameController.getVisibleEditorFrame();

		// focus visible editor frame
		if (_visibleEditorFrame) {
			_visibleEditorFrame.focus();
		}

		// if cmd equals "reload_editpage" and there are parameters, attach them to the url
		if (args[0] === "reload_editpage" || args[0] === "reload_hot_editpage") {
			url += (_currentEditorRootFrame.parameters ? _currentEditorRootFrame.parameters : '') +
				(args[1] ? '#f' + args[1] : '');
		} else if (args[0] === "remove_image" && args[2]) {
			url += '#f' + args[2];
		}

		if (_currentEditorRootFrame) {
			if (!WE().util.we_sbmtFrm(_visibleEditorFrame, url, _visibleEditorFrame)) {
				if (args[0] !== "update_image") {
					// add we_transaction, if not set
					if (!args[2]) {
						args[2] = WE().layout.weEditorFrameController.getActiveEditorFrame().getEditorTransaction();
					}
					url += "&we_transaction=" + args[2];
				}
				top.we_repl(_visibleEditorFrame, url);
			}
		}
	}
};
